import datetime as dt

from django.conf import settings
from django.contrib import messages
from django.db import connections, models
from django.db.models import F, Q, Sum
from django.db.models.expressions import RawSQL
from django.db.models.functions import Coalesce
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.audit.models import LoggedAction
from workbench.contacts.models import Organization, Person
from workbench.invoices.utils import recurring
from workbench.projects.models import Project
from workbench.services.models import ServiceBase
from workbench.tools.formats import Z1, Z2, local_date_format
from workbench.tools.models import ModelWithTotal, MoneyField, SearchQuerySet
from workbench.tools.urls import model_urls
from workbench.tools.validation import in_days, raise_if_errors


class InvoiceQuerySet(SearchQuerySet):
    def open(self):
        return self.filter(status__in=(Invoice.IN_PREPARATION, Invoice.SENT))

    def invoiced(self):
        return self.filter(status__in=Invoice.INVOICED_STATUSES)

    def overdue(self):
        return self.filter(
            status=Invoice.SENT, due_on__isnull=False, due_on__lte=in_days(-15)
        )


@model_urls
class Invoice(ModelWithTotal):
    IN_PREPARATION = 10
    SENT = 20
    PAID = 40
    CANCELED = 50

    STATUS_CHOICES = (
        (IN_PREPARATION, _("In preparation")),
        (SENT, _("Sent")),
        (PAID, _("Paid")),
        (CANCELED, _("Canceled")),
    )
    INVOICED_STATUSES = {SENT, PAID}

    FIXED = "fixed"
    DOWN_PAYMENT = "down-payment"
    SERVICES = "services"

    TYPE_CHOICES = (
        (FIXED, _("Fixed amount")),
        (DOWN_PAYMENT, _("Down payment")),
        (SERVICES, _("Services")),
    )

    customer = models.ForeignKey(
        Organization, on_delete=models.PROTECT, verbose_name=_("customer")
    )
    contact = models.ForeignKey(
        Person,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("contact"),
    )
    project = models.ForeignKey(
        Project,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("project"),
        related_name="invoices",
    )

    invoiced_on = models.DateField(_("invoiced on"), blank=True, null=True)
    due_on = models.DateField(_("due on"), blank=True, null=True)
    closed_on = models.DateField(
        _("closed on"),
        blank=True,
        null=True,
        help_text=_(
            "Payment date for paid invoices, date of"
            " replacement or cancellation otherwise."
        ),
    )
    last_reminded_on = models.DateField(_("last reminded on"), blank=True, null=True)

    title = models.CharField(_("title"), max_length=200)
    description = models.TextField(_("description"), blank=True)
    service_period_from = models.DateField(
        _("service period from"), blank=True, null=True
    )
    service_period_until = models.DateField(
        _("service period until"), blank=True, null=True
    )
    owned_by = models.ForeignKey(
        User, on_delete=models.PROTECT, verbose_name=_("responsible")
    )

    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    status = models.PositiveIntegerField(
        _("status"), choices=STATUS_CHOICES, default=IN_PREPARATION
    )
    type = models.CharField(_("type"), max_length=20, choices=TYPE_CHOICES)
    down_payment_applied_to = models.ForeignKey(
        "self",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("down payment applied to"),
        related_name="down_payment_invoices",
    )
    down_payment_total = MoneyField(_("down payment total"), default=Z2)
    third_party_costs = MoneyField(
        _("third party costs"),
        default=Z2,
        help_text=_("Only used for statistical purposes."),
    )

    postal_address = models.TextField(_("postal address"))
    _code = models.IntegerField(_("code"))
    _fts = models.TextField(editable=False, blank=True)

    payment_notice = models.TextField(
        _("payment notice"),
        blank=True,
        help_text=_("This fields' value is overridden when processing credit entries."),
    )

    objects = InvoiceQuerySet.as_manager()

    class Meta:
        ordering = ("-id",)
        verbose_name = _("invoice")
        verbose_name_plural = _("invoices")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._orig_status = self.status

    def __str__(self):
        return "%s %s - %s" % (self.code, self.title, self.owned_by.get_short_name())

    def __html__(self):
        return format_html(
            "<small>{}</small> {} - {}",
            self.code,
            self.title,
            self.owned_by.get_short_name(),
        )

    def save(self, *args, **kwargs):
        new = not self.pk
        if new:
            if self.project_id:
                self._code = RawSQL(
                    "SELECT COALESCE(MAX(_code), 0) + 1 FROM invoices_invoice"
                    " WHERE project_id = %s",
                    (self.project_id,),
                )
            else:
                self._code = RawSQL(
                    "SELECT COALESCE(MAX(_code), 0) + 1 FROM invoices_invoice"
                    " WHERE project_id IS NULL",
                    (),
                )
            super().save(*args, **kwargs)
            self.refresh_from_db()

        self._fts = " ".join(
            str(part)
            for part in [
                self.code,
                self.customer.name,
                self.contact.full_name if self.contact else "",
                self.project.title if self.project else "",
            ]
        )
        if (
            self.invoiced_on
            and self.last_reminded_on
            and self.last_reminded_on < self.invoiced_on
        ):
            # Reset last_reminded_on if it is before invoiced_on
            self.last_reminded_on = None

        if new:
            super().save()
        else:
            super().save(*args, **kwargs)

    save.alters_data = True

    def delete(self, *args, **kwargs):
        assert (
            self.status <= self.IN_PREPARATION
        ), "Trying to delete an invoice not in preparation"
        super().delete(*args, **kwargs)

    delete.alters_data = True

    def _calculate_total(self):
        if self.type == self.SERVICES:
            services = self.services.all()
            self.subtotal = sum((service.service_cost for service in services), Z2)
            self.third_party_costs = sum(
                (
                    service.third_party_costs
                    for service in services
                    if service.third_party_costs
                ),
                Z2,
            )
        super()._calculate_total()

    @property
    def code(self):
        return (
            "%s-%04d" % (self.project.code, self._code)
            if self.project
            else "%05d" % self._code
        )

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude=exclude)
        errors = {}

        if self.status >= self.SENT:
            if not self.invoiced_on or not self.due_on:
                errors["status"] = _(
                    "Invoice and/or due date missing for selected state."
                )

        if self.status <= self.SENT and self.closed_on:
            errors["status"] = _("Invalid status when closed on is already set.")

        if self.invoiced_on and self.due_on:
            if self.invoiced_on > self.due_on:
                errors["due_on"] = _("Due date has to be after invoice date.")

        if self.type in (self.SERVICES, self.DOWN_PAYMENT) and not self.project:
            errors["__all__"] = _("Invoices of type %(type)s require a project.") % {
                "type": self.get_type_display()
            }

        if self.status == Invoice.CANCELED and not self.payment_notice:
            errors["payment_notice"] = _(
                "Please provide a short reason for the invoice cancellation."
            )

        if bool(self.service_period_from) != bool(self.service_period_until):
            errors["service_period_from"] = errors["service_period_until"] = _(
                "Either fill in both fields or none."
            )
        if (
            self.service_period_from
            and self.service_period_until
            and self.service_period_from > self.service_period_until
        ):
            errors["service_period_until"] = _("Until date has to be after from date.")

        raise_if_errors(errors, exclude)

    @property
    def is_invoiced(self):
        return self.status in self.INVOICED_STATUSES

    @property
    def pretty_status(self):
        d = {
            "invoiced_on": (
                local_date_format(self.invoiced_on) if self.invoiced_on else None
            ),
            "reminded_on": (
                local_date_format(self.last_reminded_on)
                if self.last_reminded_on
                else None
            ),
            "created_at": local_date_format(self.created_at.date()),
            "closed_on": (
                local_date_format(self.closed_on) if self.closed_on else None
            ),
        }

        if self.status == self.IN_PREPARATION:
            return _("In preparation since %(created_at)s") % d
        elif self.status == self.SENT:
            if self.last_reminded_on:
                return _("Sent on %(invoiced_on)s, reminded on %(reminded_on)s") % d

            if self.due_on and dt.date.today() > self.due_on:
                return _("Sent on %(invoiced_on)s but overdue") % d

            return _("Sent on %(invoiced_on)s") % d
        elif self.status == self.PAID:
            return _("Paid on %(closed_on)s") % d
        else:
            return self.get_status_display()

    def payment_reminders_sent_at(self):
        from workbench.tools.history import changes  # Avoid a circular import

        actions = LoggedAction.objects.for_model(self).with_data(id=self.id)
        return [
            day
            for day in [
                change.values["last_reminded_on"]
                for change in changes(self, {"last_reminded_on"}, actions)
            ]
            if day
        ]

    @property
    def status_badge(self):
        css = {
            self.IN_PREPARATION: "info",
            self.SENT: "success",
            self.PAID: "default",
            self.CANCELED: "danger",
        }[self.status]

        if self.status == self.SENT:
            if self.last_reminded_on or (self.due_on and dt.date.today() > self.due_on):
                css = "warning"

        return format_html(
            '<span class="badge badge-{}">{}</span>', css, self.pretty_status
        )

    @property
    def total_title(self):
        if self.type == self.DOWN_PAYMENT:
            return (
                _("Down payment total %(currency)s incl. tax")
                if self.liable_to_vat
                else _("Down payment total %(currency)s")
            ) % {"currency": settings.WORKBENCH.CURRENCY}
        else:
            return (
                _("Total %(currency)s incl. tax")
                if self.liable_to_vat
                else _("Total %(currency)s")
            ) % {"currency": settings.WORKBENCH.CURRENCY}

    def create_services_from_logbook(self, project_services):
        assert self.project, "cannot call create_services_from_logbook without project"

        for ps in project_services:
            not_archived_effort = ps.loggedhours.filter(
                archived_at__isnull=True
            ).order_by()
            not_archived_costs = ps.loggedcosts.filter(
                archived_at__isnull=True
            ).order_by()

            hours = not_archived_effort.aggregate(Sum("hours"))["hours__sum"] or Z1
            cost = not_archived_costs.aggregate(Sum("cost"))["cost__sum"] or Z2

            if hours or cost:
                service = Service(
                    invoice=self,
                    project_service=ps,
                    title=ps.title,
                    description=ps.description,
                    position=ps.position,
                    effort_rate=ps.effort_rate,
                    effort_type=ps.effort_type,
                    effort_hours=hours,
                    cost=cost,
                    third_party_costs=not_archived_costs.filter(
                        third_party_costs__isnull=False
                    ).aggregate(Sum("third_party_costs"))["third_party_costs__sum"],
                )
                service.save(skip_related_model=True)
                not_archived_effort.update(
                    invoice_service=service, archived_at=timezone.now()
                )
                not_archived_costs.update(
                    invoice_service=service, archived_at=timezone.now()
                )

        (
            self.service_period_from,
            self.service_period_until,
        ) = self.service_period_from_logbook()
        self.save()

    def create_services_from_offer(self, project_services):
        assert self.project, "cannot call create_services_from_offer without project"

        for ps in project_services:
            service = Service(
                invoice=self,
                project_service=ps,
                title=ps.title,
                description=ps.description,
                position=ps.position,
                effort_rate=ps.effort_rate,
                effort_type=ps.effort_type,
                effort_hours=ps.effort_hours,
                cost=ps.cost,
                third_party_costs=ps.third_party_costs,
            )
            service.save(skip_related_model=True)
            ps.loggedhours.filter(archived_at__isnull=True).update(
                invoice_service=service, archived_at=timezone.now()
            )
            ps.loggedcosts.filter(archived_at__isnull=True).update(
                invoice_service=service, archived_at=timezone.now()
            )

        (
            self.service_period_from,
            self.service_period_until,
        ) = self.service_period_from_logbook()
        self.save()

    @classmethod
    def allow_delete(cls, instance, request):
        if instance.status > instance.IN_PREPARATION:
            messages.error(
                request, _("Invoices in preparation may be deleted, others not.")
            )
            return False
        return None

    def service_period_from_logbook(self):
        with connections["default"].cursor() as cursor:
            cursor.execute(
                """
with sq as (
    select min(rendered_on) as min_date, max(rendered_on) as max_date
    from logbook_loggedhours log
    left join invoices_service i_s on log.invoice_service_id=i_s.id
    where i_s.invoice_id=%s

    union all

    select min(rendered_on) as min_date, max(rendered_on) as max_date
    from logbook_loggedcost log
    left join invoices_service i_s on log.invoice_service_id=i_s.id
    where i_s.invoice_id=%s
)
select min(min_date), max(max_date) from sq
                """,
                [self.id, self.id],
            )
            return list(cursor)[0]

    @property
    def service_period(self):
        period = [self.service_period_from, self.service_period_until]
        return (
            "%s - %s" % tuple(local_date_format(day) for day in period)
            if all(period)
            else None
        )


@model_urls
class Service(ServiceBase):
    RELATED_MODEL_FIELD = "invoice"

    invoice = models.ForeignKey(
        Invoice,
        on_delete=models.CASCADE,
        related_name="services",
        verbose_name=_("invoice"),
    )
    project_service = models.ForeignKey(
        "projects.Service",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="invoice_services",
        verbose_name=_("project service"),
    )

    def get_absolute_url(self):
        return self.invoice.get_absolute_url()

    @classmethod
    def allow_update(cls, instance, request):
        if instance.invoice.status > instance.invoice.IN_PREPARATION:
            messages.error(
                request,
                _(
                    "Cannot modify a service bound to an invoice"
                    " which is not in preparation anymore."
                ),
            )
            return False
        return True

    allow_delete = allow_update

    @classmethod
    def get_redirect_url(cls, instance, request):
        if not request.is_ajax():
            return instance.get_absolute_url() if instance else "invoices_invoice_list"


class RecurringInvoiceQuerySet(SearchQuerySet):
    def renewal_candidates(self):
        today = dt.date.today()
        return (
            self.annotate(_start=Coalesce("next_period_starts_on", "starts_on"))
            .filter(
                Q(_start__lte=today - F("create_invoice_on_day")),
                Q(ends_on__isnull=True) | Q(ends_on__gte=F("_start")),
            )
            .select_related("customer", "contact", "owned_by")
        )


@model_urls
class RecurringInvoice(ModelWithTotal):
    PERIODICITY_CHOICES = [
        ("yearly", _("yearly")),
        ("quarterly", _("quarterly")),
        ("monthly", _("monthly")),
        ("weekly", _("weekly")),
    ]

    customer = models.ForeignKey(
        Organization, on_delete=models.PROTECT, verbose_name=_("customer")
    )
    contact = models.ForeignKey(
        Person,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("contact"),
    )

    title = models.CharField(_("title"), max_length=200)
    description = models.TextField(_("description"), blank=True)
    owned_by = models.ForeignKey(
        User, on_delete=models.PROTECT, verbose_name=_("responsible")
    )

    created_at = models.DateTimeField(_("created at"), default=timezone.now)

    third_party_costs = MoneyField(
        _("third party costs"),
        default=Z2,
        help_text=_("Only used for statistical purposes."),
    )
    postal_address = models.TextField(_("postal address"))

    starts_on = models.DateField(_("starts on"), default=dt.date.today)
    ends_on = models.DateField(_("ends on"), blank=True, null=True)
    periodicity = models.CharField(
        _("periodicity"), max_length=20, choices=PERIODICITY_CHOICES
    )
    create_invoice_on_day = models.IntegerField(
        _("create invoice on day"),
        default=-20,
        help_text=_(
            "Invoices are created 20 days before their period begins by default."
        ),
    )
    next_period_starts_on = models.DateField(
        _("next period starts on"), blank=True, null=True
    )

    create_project = models.BooleanField(
        _("Create project?"),
        help_text=_("Invoices are created without projects by default."),
        default=False,
    )

    objects = RecurringInvoiceQuerySet.as_manager()

    class Meta:
        ordering = ["customer__name", "title"]
        verbose_name = _("recurring invoice")
        verbose_name_plural = _("recurring invoices")

    def __str__(self):
        return self.title

    def __html__(self):
        return "%s - %s" % (self.title, self.owned_by.get_short_name())

    @property
    def pretty_status(self):
        if self.ends_on:
            return _("%(periodicity)s from %(from)s until %(until)s") % {
                "periodicity": self.get_periodicity_display(),
                "from": local_date_format(self.starts_on),
                "until": local_date_format(self.ends_on),
            }
        return _("%(periodicity)s from %(from)s") % {
            "periodicity": self.get_periodicity_display(),
            "from": local_date_format(self.starts_on),
        }

    @property
    def pretty_next_period(self):
        start = self.next_period_starts_on or self.starts_on
        if self.ends_on and self.ends_on < start:
            return ""
        create = start + dt.timedelta(days=self.create_invoice_on_day)

        return _(
            "Next period starts on %(start)s, invoice will be created on %(create)s"
        ) % {"start": local_date_format(start), "create": local_date_format(create)}

    @property
    def status_badge(self):
        return format_html(
            '<span class="badge badge-{}">{}</span>',
            "light" if self.ends_on else "secondary",
            self.pretty_status,
        )

    def create_single_invoice(self, *, period_starts_on, period_ends_on):
        project = None
        if self.create_project:
            project = Project.objects.create(
                customer=self.customer,
                contact=self.contact,
                title=self.title,
                description=self.description,
                owned_by=self.owned_by,
                type=Project.MAINTENANCE,
            )

        return Invoice.objects.create(
            customer=self.customer,
            contact=self.contact,
            project=project,
            invoiced_on=period_starts_on,
            due_on=period_starts_on + dt.timedelta(days=15),
            title=self.title,
            description=self.description,
            service_period_from=period_starts_on,
            service_period_until=period_ends_on,
            owned_by=self.owned_by,
            status=Invoice.IN_PREPARATION,
            type=Invoice.DOWN_PAYMENT if self.create_project else Invoice.FIXED,
            postal_address=self.postal_address,
            subtotal=self.subtotal,
            discount=self.discount,
            liable_to_vat=self.liable_to_vat,
            # tax_rate=self.tax_rate,
            # total=self.total,
            third_party_costs=self.third_party_costs,
        )

    def create_invoices(self):
        invoices = []
        days = recurring(
            max(filter(None, (self.next_period_starts_on, self.starts_on))),
            self.periodicity,
        )
        generate_until = min(
            filter(None, (in_days(-self.create_invoice_on_day), self.ends_on))
        )
        this_period = next(days)
        while True:
            if this_period > generate_until:
                break
            next_period = next(days)
            invoices.append(
                self.create_single_invoice(
                    period_starts_on=this_period,
                    period_ends_on=next_period - dt.timedelta(days=1),
                )
            )
            self.next_period_starts_on = next_period
            this_period = next_period
        self.save()
        return invoices



from django.db import models
from django.utils.translation import gettext_lazy as _

from workbench.projects.models import Project
from workbench.reporting.project_budget_statistics import project_budget_statistics
from workbench.tools.formats import local_date_format
from workbench.tools.models import MoneyField


class AccrualsQuerySet(models.QuerySet):
    def accruals(self, cutoff_date):
        projects = Project.objects.open(on=cutoff_date)
        statistics = project_budget_statistics(projects, cutoff_date=cutoff_date)
        return statistics["overall"]["delta_negative"]

    def for_cutoff_date(self, cutoff_date):
        instance, created = self.update_or_create(
            cutoff_date=cutoff_date,
            defaults={"accruals": self.accruals(cutoff_date)},
        )
        return instance


class Accruals(models.Model):
    cutoff_date = models.DateField(_("cutoff date"), unique=True)
    accruals = MoneyField(_("accruals"))

    objects = AccrualsQuerySet.as_manager()

    class Meta:
        ordering = ["-cutoff_date"]
        verbose_name = _("accruals")
        verbose_name_plural = _("accruals")

    def __str__(self):
        return local_date_format(self.cutoff_date)


class CostCenter(models.Model):
    title = models.CharField(_("title"), max_length=200)
    position = models.PositiveIntegerField(_("position"), default=0)

    class Meta:
        ordering = ("position", "id")
        verbose_name = _("cost center")
        verbose_name_plural = _("cost centers")

    def __str__(self):
        return self.title



# Generated by Django 2.2.1 on 2019-05-18 12:25

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("logbook", "0006_auto_20190321_1448"),
    ]

    operations = [
        migrations.AddField(
            model_name="loggedcost",
            name="are_expenses",
            field=models.BooleanField(
                default=False, verbose_name="paid from my own pocket"
            ),
        ),
        migrations.AddField(
            model_name="loggedcost",
            name="expenses_reimbursed_at",
            field=models.DateTimeField(
                blank=True, null=True, verbose_name="expenses reimbursed at"
            ),
        ),
        migrations.AddField(
            model_name="loggedcost",
            name="rendered_by",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="loggedcosts",
                to=settings.AUTH_USER_MODEL,
                verbose_name="rendered by",
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2rc1 on 2019-03-24 17:15

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("credit_control", "0004_auto_20190324_1813")]

    operations = [
        migrations.AlterField(
            model_name="creditentry",
            name="ledger",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="transactions",
                to="credit_control.Ledger",
                verbose_name="ledger",
            ),
            preserve_default=False,
        )
    ]



from workbench.tools import admin

from . import models


@admin.register(models.Offer)
class OfferAdmin(admin.ModelAdmin):
    list_display = ("project", "title", "offered_on", "owned_by", "status", "total")
    list_filter = ("status",)
    raw_id_fields = ("project",)



# Generated by Django 2.2.4 on 2019-08-18 13:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("projects", "0013_auto_20190630_2219")]

    operations = [
        migrations.RunSQL(
            "UPDATE projects_project SET type='order' WHERE type='acquisition'"
        ),
        migrations.AlterField(
            model_name="project",
            name="type",
            field=models.CharField(
                choices=[
                    ("order", "Order"),
                    ("maintenance", "Maintenance"),
                    ("internal", "Internal"),
                ],
                max_length=20,
                verbose_name="type",
            ),
        ),
    ]



from django import template


register = template.Library()


@register.tag
def formset(parser, token):
    """
    Implements formsets where subforms can be added using the
    ``towel_add_subform`` javascript method::

        {% formset formset "activities" %}
            ... form code
        {% endformset %}
    """

    tokens = token.split_contents()
    nodelist = parser.parse(("endformset",))
    parser.delete_first_token()

    return DynamicFormsetNode(tokens[1], tokens[2], nodelist)


class DynamicFormsetNode(template.Node):
    def __init__(self, formset, slug, nodelist):
        self.formset = template.Variable(formset)
        self.slug = template.Variable(slug)
        self.nodelist = nodelist

    def render(self, context):
        formset = self.formset.resolve(context)
        slug = self.slug.resolve(context)

        result = [str(formset.management_form)]

        with context.push(form_id="%s-empty" % slug, form=formset.empty_form):
            result.append('<script type="text/template" id="%s-empty">' % slug)
            result.append(self.nodelist.render(context))
            result.append("</script>")

        for idx, form in enumerate(formset.forms):
            with context.push(form_id="%s-%s" % (slug, idx), form=form):
                result.append(self.nodelist.render(context))

        return "".join(result)



from django import forms
from django.template.loader import render_to_string

from fineforms.wrappers import FieldWrapper, html_safe


@html_safe
class BootstrapFieldWrapper(FieldWrapper):
    template_name = "fineforms/field.html"
    label_suffix = ""
    error_css_class = "is-invalid"
    required_css_class = "required"

    def __init__(self, field):
        self.field = field

    def __str__(self):
        extra_classes = []
        if not hasattr(self.field.form, "error_css_class") and self.field.errors:
            extra_classes.append(self.error_css_class)
        if (
            not hasattr(self.field.form, "required_css_class")
            and self.field.field.required
        ):
            extra_classes.append(self.required_css_class)

        widget = self.field.field.widget
        extra_label_classes = []
        if isinstance(widget, forms.CheckboxInput):
            widget.attrs["class"] = widget.attrs.get("class", "") + " form-check-input"
            extra_label_classes.append("form-check-label")
        elif isinstance(widget, (forms.RadioSelect, forms.CheckboxSelectMultiple)):
            widget.attrs["class"] = " ".join(
                filter(None, (widget.attrs.get("class", ""), "my-2"))
            )
        else:
            cls = "form-control"
            if isinstance(widget, forms.Select):
                cls = "custom-select"
            elif isinstance(widget, forms.FileInput):
                cls = "form-control-file"
            widget.attrs["class"] = " ".join(
                filter(
                    None,
                    (
                        widget.attrs.get("class", ""),
                        cls,
                        " is-invalid" if self.field.errors else "",
                    ),
                )
            )

        html = render_to_string(
            self.template_name,
            {
                "field": self.field,
                "widget_then_label": isinstance(
                    self.field.field.widget, forms.CheckboxInput
                ),
                "label_tag": self.field.label_tag(
                    label_suffix=self.label_suffix,
                    attrs=({"class": " ".join(extra_classes + extra_label_classes)}),
                ),
                "css_classes": self.field.css_classes(
                    extra_classes=extra_classes
                    + [
                        "widget--%s"
                        % (self.field.field.widget.__class__.__name__.lower(),)
                    ]
                ),
            },
        )

        if isinstance(widget, forms.RadioSelect):
            html = html.replace('type="radio"', 'class="form-check-input" type="radio"')
        elif isinstance(widget, forms.CheckboxSelectMultiple):
            html = html.replace(
                'type="checkbox"', 'class="form-check-input" type="checkbox"'
            )
        return html



# Generated by Django 2.2.5 on 2019-09-13 07:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contacts", "0010_organization_default_billing_address")]

    operations = [
        migrations.AlterField(
            model_name="organization",
            name="groups",
            field=models.ManyToManyField(
                blank=True, to="contacts.Group", verbose_name="groups"
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="groups",
            field=models.ManyToManyField(
                blank=True, to="contacts.Group", verbose_name="groups"
            ),
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-05 07:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contacts", "0002_auto_20190304_2247")]

    operations = [
        migrations.AddField(
            model_name="person",
            name="salutation",
            field=models.CharField(
                help_text="Dear John/Dear Ms Smith",
                max_length=100,
                verbose_name="complete salutation",
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="address",
            field=models.CharField(
                blank=True, help_text="Mr./Ms.", max_length=100, verbose_name="address"
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-15 10:22

from django.db import migrations

from workbench.accounts.middleware import set_user_name


def forwards(apps, schema_editor):
    set_user_name("Migrations")
    LoggedCost = apps.get_model("logbook", "LoggedCost")
    Service = apps.get_model("projects", "Service")

    project_ids_with_costs_without_services = set(
        LoggedCost.objects.filter(service_id__isnull=True).values_list(
            "project_id", flat=True
        )
    )

    for project_id in project_ids_with_costs_without_services:
        service = Service.objects.create(
            project_id=project_id,
            title="Keine Leistung",
            position=-1,
            allow_logging=False,
        )
        LoggedCost.objects.filter(project=project_id, service__isnull=True).update(
            service=service
        )


class Migration(migrations.Migration):

    dependencies = [
        ("logbook", "0012_auto_20190913_0919"),
        ("projects", "0016_auto_20190915_1249"),
    ]

    operations = [migrations.RunPython(forwards)]



# Generated by Django 3.0 on 2019-12-10 14:54

import django.db.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("expenses", "0003_auto_20190719_1354"),
    ]

    operations = [
        migrations.CreateModel(
            name="ExchangeRates",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("day", models.DateField(unique=True, verbose_name="day")),
                ("rates", django.db.models.JSONField()),
            ],
            options={
                "verbose_name": "exchange rate",
                "verbose_name_plural": "exchange rates",
                "ordering": ["-day"],
                "get_latest_by": "day",
            },
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-04 21:39

from decimal import Decimal

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ServiceType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=40, verbose_name="title")),
                (
                    "billing_per_hour",
                    models.DecimalField(
                        decimal_places=2,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="billing per hour",
                    ),
                ),
                (
                    "position",
                    models.PositiveIntegerField(default=0, verbose_name="position"),
                ),
            ],
            options={
                "verbose_name": "service type",
                "verbose_name_plural": "service types",
                "ordering": ("position", "id"),
            },
        )
    ]



import datetime as dt
import json
import os
from unittest import mock

from django.conf import settings
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone

from workbench import factories
from workbench.expenses.models import ExchangeRates, ExpenseReport
from workbench.expenses.rates import exchange_rates
from workbench.logbook.models import LoggedCost
from workbench.tools.formats import local_date_format
from workbench.tools.testing import check_code, messages
from workbench.tools.validation import in_days


class ExpensesTest(TestCase):
    fixtures = ["exchangerates.json"]

    def test_logged_cost_deletion(self):
        """Archived logged costs cannot be deleted, others can"""
        costs = factories.LoggedCostFactory.create(archived_at=timezone.now())
        self.client.force_login(costs.created_by)

        response = self.client.post(
            costs.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            messages(response), ["Cannot delete archived logged cost entries."]
        )

        costs = factories.LoggedCostFactory.create()
        response = self.client.post(
            costs.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 204)

    def test_list(self):
        """Filter form smoke test"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        report = ExpenseReport.objects.create(created_by=user, owned_by=user)
        report.expenses.set(
            [
                factories.LoggedCostFactory.create(
                    rendered_by=user, are_expenses=True, third_party_costs=5
                )
            ]
        )

        code = check_code(self, "/expenses/")
        code("")
        code("s=in-preparation")
        code("s=closed")
        code("owned_by={}".format(user.id))
        code("owned_by=-1")  # mine
        code("owned_by=0")  # only inactive

    def test_expenses(self):
        """Expense logging and expense report creation"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        service = factories.ServiceFactory.create(project=project)

        response = self.client.post(
            project.urls["createcost"],
            {
                "modal-service": service.id,
                "modal-rendered_by": project.owned_by_id,
                "modal-rendered_on": local_date_format(dt.date.today()),
                "modal-cost": "10",
                "modal-description": "Anything",
                "modal-are_expenses": "on",
                # "modal-third_party_costs": "9",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(
            response, "Providing third party costs is necessary for expenses."
        )

        response = self.client.post(
            project.urls["createcost"],
            {
                "modal-service": service.id,
                "modal-rendered_by": project.owned_by_id,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-cost": "10",
                "modal-description": "Anything",
                "modal-are_expenses": "on",
                "modal-third_party_costs": "9",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        cost1 = LoggedCost.objects.get()
        self.assertEqual(
            list(LoggedCost.objects.expenses(user=project.owned_by)), [cost1]
        )

        # Another users' expenses
        factories.LoggedCostFactory.create(are_expenses=True, third_party_costs=5)

        response = self.client.get("/expenses/create/")
        # Only one expense
        self.assertContains(response, "id_expenses_0")
        self.assertNotContains(response, "id_expenses_1")

        response = self.client.post("/expenses/create/", {"expenses": [cost1.pk]})
        report = ExpenseReport.objects.get()
        self.assertRedirects(response, report.urls["detail"])
        self.assertEqual(report.total, 9)

        # Some fields are now disabled
        response = self.client.get(
            cost1.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response,
            '<input type="number" name="modal-third_party_costs" value="9.00" step="0.01" class="form-control" disabled id="id_modal-third_party_costs">',  # noqa
            html=True,
        )

        response = self.client.get(
            cost1.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response, "Expenses are part of an expense report, cannot delete entry."
        )

        # More expenses of same user
        cost2 = factories.LoggedCostFactory.create(
            rendered_by=project.owned_by, are_expenses=True, third_party_costs=5
        )
        response = self.client.post(report.urls["update"], {"expenses": [cost2.pk]})
        self.assertRedirects(response, report.urls["detail"])
        report.refresh_from_db()
        self.assertEqual(report.total, 5)

        response = self.client.get(report.urls["pdf"])
        self.assertRedirects(response, report.urls["detail"])
        self.assertEqual(
            messages(response),
            [
                "Please close the expense report first. Generating PDFs"
                " for open expense reports isn't allowed."
            ],
        )

        response = self.client.post(report.urls["delete"])
        self.assertRedirects(response, "/expenses/")

        self.assertEqual(ExpenseReport.objects.count(), 0)

        response = self.client.post(
            "/expenses/create/", {"expenses": [cost1.pk], "is_closed": "on"}
        )
        report = ExpenseReport.objects.get()
        self.assertRedirects(response, report.urls["detail"])
        self.assertEqual(report.total, 9)

        response = self.client.post(report.urls["update"])
        self.assertRedirects(response, report.urls["detail"])
        self.assertEqual(messages(response), ["Cannot update a closed expense report."])

        response = self.client.post(report.urls["delete"])
        self.assertRedirects(response, report.urls["detail"])
        self.assertEqual(messages(response), ["Cannot delete a closed expense report."])
        # print(response, response.content.decode("utf-8"))

        self.assertEqual(self.client.get(report.urls["pdf"]).status_code, 200)

    def test_no_expenses(self):
        """Expense report creation without expenses fails early"""
        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/expenses/create/")
        self.assertRedirects(response, "/expenses/")
        self.assertEqual(
            messages(response), ["Could not find any expenses to reimburse."]
        )

    def test_exchange_rate_str(self):
        """__str__ method of exchange rates..."""
        today = dt.date.today()
        self.assertEqual(str(ExchangeRates(day=today)), str(today))

    def test_exchange_rate_conversion(self):
        """The exchange rate conversion API returns expected values"""
        # From https://api.exchangeratesapi.io/latest?base=CHF, but pruned
        ExchangeRates.objects.create(
            day=dt.date(2019, 12, 11),
            rates={
                "rates": {
                    "CHF": 1.0,
                    "EUR": 0.9160864786,
                    "USD": 1.014565775,
                    "PLN": 3.927171125,
                },
                "base": "CHF",
                "date": "2019-12-11",
            },
        )

        self.client.force_login(factories.UserFactory.create())

        response = self.client.get("/expenses/convert/")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"cost": ""})

        response = self.client.get(
            "/expenses/convert/?day=2019-12-11&currency=EUR&cost=100"
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"cost": "109.16"})

    def test_one_or_other(self):
        """Foreign expenses require both the expense amount and the currency"""
        service = factories.ServiceFactory.create()
        kw = {
            "service": service,
            "created_by": service.project.owned_by,
            "rendered_by": service.project.owned_by,
            "cost": 10,
            "description": "Test",
        }
        with self.assertRaises(ValidationError) as cm:
            LoggedCost(**kw, expense_currency="EUR").full_clean()
        self.assertEqual(
            list(cm.exception),
            [("expense_cost", ["Either fill in all fields or none."])],
        )

        with self.assertRaises(ValidationError) as cm:
            LoggedCost(**kw, expense_cost=10).full_clean()
        self.assertEqual(
            list(cm.exception),
            [("expense_currency", ["Either fill in all fields or none."])],
        )


def mocked_json(*args, **kwargs):
    with open(
        os.path.join(settings.BASE_DIR, "workbench", "fixtures", "exchangerates.json")
    ) as f:
        return json.load(f)[0]["fields"]["rates"]


@mock.patch("workbench.expenses.models.exchange_rates", side_effect=mocked_json)
class MockedRemoteDataTest(TestCase):
    def test_with_mocked_remote_data(self, mock_get):
        """Exchange rates determination"""
        self.assertEqual(mock_get.call_count, 0)
        rates = ExchangeRates.objects.newest()
        self.assertEqual(mock_get.call_count, 1)
        self.assertEqual(rates.rates["date"], "2019-12-10")

        rates = ExchangeRates.objects.create(day=in_days(1), rates={})
        self.assertEqual(ExchangeRates.objects.newest(), rates)
        self.assertEqual(mock_get.call_count, 1)


def mocked_get(*args, **kwargs):
    class MockRequest:
        def json(self):
            return {}

    return MockRequest()


@mock.patch("workbench.expenses.rates.requests.get", side_effect=mocked_get)
class ExchangeRatesTest(TestCase):
    def test_exchange_rates_today(self, mock_get):
        """exchange_rates() without arguments fetches today's exchange rates"""
        exchange_rates()
        self.assertEqual(
            mock_get.call_args[0],
            ("https://api.exchangeratesapi.io/latest?base=CHF",),
        )

    def test_exchange_rates_someday(self, mock_get):
        """Exchange rates of a specific day hits the expected URL"""
        exchange_rates(dt.date(2019, 10, 12))
        self.assertEqual(
            mock_get.call_args[0],
            ("https://api.exchangeratesapi.io/2019-10-12?base=CHF",),
        )



from django.urls import re_path

from workbench import generic
from workbench.accounts.features import bookkeeping_only, controlling_only
from workbench.invoices import views
from workbench.invoices.forms import (
    CreatePersonInvoiceForm,
    InvoiceDeleteForm,
    InvoiceForm,
    InvoiceSearchForm,
    ServiceForm,
)
from workbench.invoices.models import Invoice, Service


urlpatterns = [
    re_path(
        r"^$",
        controlling_only(
            generic.ListView.as_view(model=Invoice, search_form_class=InvoiceSearchForm)
        ),
        name="invoices_invoice_list",
    ),
    re_path(
        r"^autocomplete/$",
        generic.AutocompleteView.as_view(
            model=Invoice,
            queryset=Invoice.objects.select_related("project", "owned_by"),
        ),
        name="invoices_invoice_autocomplete",
    ),
    re_path(
        r"^(?P<pk>\d+)/$",
        controlling_only(generic.DetailView.as_view(model=Invoice)),
        name="invoices_invoice_detail",
    ),
    re_path(
        r"^create/$",
        controlling_only(
            generic.CreateView.as_view(
                model=Invoice, form_class=CreatePersonInvoiceForm
            )
        ),
        name="invoices_invoice_create",
    ),
    re_path(
        r"^(?P<pk>\d+)/update/$",
        controlling_only(
            generic.UpdateView.as_view(form_class=InvoiceForm, model=Invoice)
        ),
        name="invoices_invoice_update",
    ),
    re_path(
        r"^(?P<pk>\d+)/delete/$",
        controlling_only(
            generic.DeleteView.as_view(
                model=Invoice, delete_form_class=InvoiceDeleteForm
            )
        ),
        name="invoices_invoice_delete",
    ),
    re_path(
        r"^(?P<pk>\d+)/pdf/$",
        controlling_only(views.InvoicePDFView.as_view()),
        name="invoices_invoice_pdf",
    ),
    re_path(
        r"^(?P<pk>\d+)/xlsx/$",
        controlling_only(views.InvoiceXLSXView.as_view()),
        name="invoices_invoice_xlsx",
    ),
    # Services
    re_path(
        r"^(?P<pk>\d+)/createservice/$",
        controlling_only(
            generic.CreateRelatedView.as_view(
                model=Service, form_class=ServiceForm, related_model=Invoice
            )
        ),
        name="invoices_invoice_createservice",
    ),
    re_path(
        r"^services/(?P<pk>[0-9]+)/update/$",
        controlling_only(
            generic.UpdateView.as_view(model=Service, form_class=ServiceForm)
        ),
        name="invoices_service_update",
    ),
    re_path(
        r"^reminders/$", bookkeeping_only(views.reminders), name="invoices_reminders"
    ),
    re_path(
        r"^dunning-letter/(?P<customer_id>[0-9]+)/$",
        bookkeeping_only(views.dunning_letter),
        name="invoices_dunning_letter",
    ),
]



# Generated by Django 3.0.4 on 2020-03-11 16:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("offers", "0008_auto_20190916_1110"),
        ("deals", "0003_valuetype_weekly_target"),
    ]

    operations = [
        migrations.AddField(
            model_name="deal",
            name="related_offers",
            field=models.ManyToManyField(
                blank=True,
                related_name="deals",
                to="offers.Offer",
                verbose_name="related offers",
            ),
        ),
    ]



from django.shortcuts import redirect, render
from django.utils.translation import gettext as _

from workbench.projects.forms import ProjectAutocompleteForm
from workbench.templatetags.workbench import h


def create(request, *, viewname):
    if not request.is_ajax():
        return redirect("/")
    form = ProjectAutocompleteForm(request.POST if request.method == "POST" else None)
    params = request.GET.urlencode()
    if form.is_valid():
        if form.cleaned_data["service"]:
            params += "&service={}".format(form.cleaned_data["service"].pk)
        return redirect(
            "{}?{}".format(form.cleaned_data["project"].urls[viewname], params)
        )
    return render(
        request,
        "generic/select_object.html",
        {
            "form": form,
            "title": _("Select project for logging"),
            "links": [
                {
                    "title": h(project),
                    "url": "{}?{}".format(project.urls[viewname], params),
                    "attrs": 'data-toggle="ajaxmodal"',
                }
                for project in request.user.active_projects
            ],
        },
    )



import datetime as dt
from collections import defaultdict

from django.utils.translation import gettext as _

from workbench.tools.reporting import query
from workbench.tools.validation import in_days, monday


def logged_hours(user):
    stats = {}

    from_ = monday(in_days(-180))

    hours_per_week = {}
    for week, type, hours in query(
        """
WITH sq AS (
    SELECT
        date_trunc('week', rendered_on) AS week,
        project.type AS type,
        SUM(hours) AS hours
    FROM logbook_loggedhours hours
    LEFT JOIN projects_service service ON hours.service_id=service.id
    LEFT JOIN projects_project project ON service.project_id=project.id
    WHERE rendered_by_id=%s AND rendered_on>=%s
    GROUP BY week, project.type
)
SELECT series.week, sq.type, COALESCE(sq.hours, 0)
FROM generate_series(%s, %s, '7 days') AS series(week)
LEFT OUTER JOIN sq ON series.week=sq.week
ORDER BY series.week
""",
        [user.id, from_, from_, monday() + dt.timedelta(days=6)],
    ):
        if week in hours_per_week:
            hours_per_week[week]["hours"] += hours
            hours_per_week[week]["by_type"][type] = hours
        else:
            hours_per_week[week] = {
                "week": week,
                "hours": hours,
                "by_type": {type: hours},
            }

    stats["hours_per_week"] = [row[1] for row in sorted(hours_per_week.items())]

    hours_per_customer = defaultdict(dict)
    total_hours_per_customer = defaultdict(int)

    for week, customer, hours in query(
        """
WITH sq AS (
    SELECT
        date_trunc('week', rendered_on) AS week,
        customer.name AS customer,
        SUM(hours) AS hours
    FROM logbook_loggedhours hours
    LEFT JOIN projects_service service ON hours.service_id=service.id
    LEFT JOIN projects_project project ON service.project_id=project.id
    LEFT JOIN contacts_organization customer ON project.customer_id=customer.id
    WHERE rendered_by_id=%s AND rendered_on>=%s
    GROUP BY week, customer.name
)
SELECT series.week, COALESCE(sq.customer, ''), COALESCE(sq.hours, 0)
FROM generate_series(%s, %s, '7 days') AS series(week)
LEFT OUTER JOIN sq ON series.week=sq.week
ORDER BY series.week
        """,
        [user.id, from_, from_, monday() + dt.timedelta(days=6)],
    ):
        customer = customer.split("\n")[0]
        hours_per_customer[week][customer] = hours
        total_hours_per_customer[customer] += hours

    customers = [
        row[0]
        for row in sorted(
            total_hours_per_customer.items(), key=lambda row: row[1], reverse=True
        )
    ][:10]

    weeks = sorted(hours_per_customer.keys())
    stats["hours_per_customer"] = {
        "weeks": weeks,
        "by_customer": [
            {
                "name": customer,
                "hours": [hours_per_customer[week].get(customer, 0) for week in weeks],
            }
            for customer in customers
        ],
    }
    customers = set(customers)
    stats["hours_per_customer"]["by_customer"].append(
        {
            "name": _("All others"),
            "hours": [
                sum(
                    (
                        hours
                        for customer, hours in hours_per_customer[week].items()
                        if customer not in customers
                    ),
                    0,
                )
                for week in weeks
            ],
        }
    )

    dows = [
        None,
        _("Monday"),
        _("Tuesday"),
        _("Wednesday"),
        _("Thursday"),
        _("Friday"),
        _("Saturday"),
        _("Sunday"),
    ]

    stats["rendered_hours_per_weekday"] = [
        {"dow": int(dow), "name": dows[int(dow)], "hours": hours}
        for dow, hours in query(
            """
WITH sq AS (
    SELECT
        (extract(isodow from rendered_on)::integer) as dow,
        SUM(hours) AS hours
    FROM logbook_loggedhours
    WHERE rendered_by_id=%s AND rendered_on>=%s
    GROUP BY dow
    ORDER BY dow
)
SELECT series.dow, COALESCE(sq.hours, 0)
FROM generate_series(1, 7) AS series(dow)
LEFT OUTER JOIN sq ON series.dow=sq.dow
ORDER BY series.dow
            """,
            [user.id, from_],
        )
    ]

    stats["created_hours_per_weekday"] = [
        {"dow": int(dow), "name": dows[int(dow)], "hours": hours}
        for dow, hours in query(
            """
WITH sq AS (
    SELECT
        (extract(isodow from timezone('CET', created_at))::integer) as dow,
        SUM(hours) AS hours
    FROM logbook_loggedhours
    WHERE rendered_by_id=%s AND rendered_on>=%s
    GROUP BY dow
    ORDER BY dow
)
SELECT series.dow, COALESCE(sq.hours, 0)
FROM generate_series(1, 7) AS series(dow)
LEFT OUTER JOIN sq ON series.dow=sq.dow
ORDER BY series.dow
            """,
            [user.id, from_],
        )
    ]

    return stats


def work_anniversaries():
    anniversaries = []
    today = dt.date.today()

    for name, start in query(
        """
WITH
employments AS (
  SELECT
    user_id,
    date_from,
    date_until
  FROM awt_employment
  WHERE percentage>0
  ORDER BY date_from
),
ends AS (
  SELECT
    user_id,
    array_agg(date_until + 1) AS ends
  FROM employments
  GROUP BY user_id
),
earliest AS (
  SELECT
    user_id,
    -- The latest start which ...
    MAX(date_from) AS start
  FROM employments
  LEFT JOIN ends USING (user_id)
  -- ... has no ending employment preceeding it (one day difference)
  WHERE date_from <> ALL(ends.ends)
  GROUP BY user_id
)
SELECT _full_name, earliest.start
FROM accounts_user u
LEFT JOIN earliest ON u.id=earliest.user_id
WHERE earliest.start IS NOT NULL AND u.is_active=TRUE
        """,
        [],
    ):
        this_year = start.replace(year=today.year)
        anniversaries.append(
            {
                "name": name,
                "start": start,
                "this_year": this_year,
                "anniversary": this_year.year - start.year,
                "already": this_year > today,
            }
        )

    return sorted(anniversaries, key=lambda row: (row["this_year"], row["name"]))


def average_employment_duration():
    stats = {}

    stats["current_users"] = query(
        """
with date_from_until as (
    select
        user_id,
        date_from,
        case
            when date_until='9999-12-31' then current_date
            else date_until
        end as date_until
    from awt_employment e
    left join accounts_user u on e.user_id=u.id
    where u.is_active=TRUE
),
durations as (
select user_id, sum(date_until - date_from) as duration
from date_from_until
group by user_id
)
select avg(duration)/365.24 from durations
        """,
        [],
    )[0][0]

    stats["all_users"] = query(
        """
with date_from_until as (
    select
        user_id,
        date_from,
        case
            when date_until='9999-12-31' then current_date
            else date_until
        end as date_until
    from awt_employment
),
durations as (
select user_id, sum(date_until - date_from) as duration
from date_from_until
group by user_id
)
select avg(duration)/365.24 from durations
        """,
        [],
    )[0][0]

    return stats


def test():  # pragma: no cover
    from pprint import pprint

    # from workbench.accounts.models import User
    # u = User.objects.get(pk=1)
    # pprint(logged_hours(u))

    pprint(work_anniversaries())



# Generated by Django 3.1.2 on 2020-10-16 07:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("accounts", "0009_user_person"),
    ]

    operations = [
        migrations.RunSQL(
            """
SELECT audit_audit_table('accounts_team');
UPDATE accounts_team SET id=id;
                """,
            "",
        ),
    ]



from decimal import Decimal
from functools import partial
from itertools import islice

from django.contrib import messages
from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import ProtectedError
from django.db.models.deletion import Collector
from django.utils.translation import gettext, gettext_lazy as _

from workbench.tools.formats import Z2, currency
from workbench.tools.search import search


class SearchQuerySet(models.QuerySet):
    def search(self, terms):
        return search(self, terms)


class SlowCollector(Collector):
    def can_fast_delete(self, *args, **kwargs):
        return False


class Model(models.Model):
    objects = SearchQuerySet.as_manager()

    class Meta:
        abstract = True

    @classmethod
    def allow_create(cls, request):
        return True

    @classmethod
    def allow_update(cls, instance, request):
        return True

    @classmethod
    def allow_delete(cls, instance, request):
        collector = SlowCollector(using=instance._state.db)
        try:
            collector.collect([instance])
        except ProtectedError as exc:
            messages.error(
                request,
                gettext(
                    "Cannot delete '%(object)s'"
                    " because of related objects (%(related)s)."
                )
                % {
                    "object": instance,
                    "related": ", ".join(
                        str(o) for o in islice(exc.protected_objects, 10)
                    ),
                },
            )
            return False
        else:
            return True

    @classmethod
    def get_redirect_url(cls, instance, request):
        return None

    @property
    def code(self):
        return "%05d" % self.pk if self.pk else ""

    @property
    def pretty_status(self):
        return self.get_status_display() if hasattr(self, "get_status_display") else ""


ONE = Decimal("1")


MoneyField = partial(
    models.DecimalField,
    max_digits=10,
    decimal_places=2,
    validators=[MinValueValidator(0)],
)


HoursField = partial(
    models.DecimalField,
    max_digits=10,
    decimal_places=1,
    validators=[MinValueValidator(0)],
)


HoursFieldAllowNegatives = partial(models.DecimalField, max_digits=10, decimal_places=1)


class ModelWithTotal(Model):
    subtotal = MoneyField(_("subtotal"), default=Z2)
    discount = MoneyField(_("discount"), default=Z2)
    liable_to_vat = models.BooleanField(
        _("liable to VAT"),
        default=True,
        help_text=_(
            "For example invoices to foreign institutions are not liable to VAT."
        ),
    )
    total_excl_tax = MoneyField(_("total excl. tax"), default=Z2)
    tax_rate = MoneyField(_("tax rate"), default=Decimal("7.7"))
    total = MoneyField(_("total"), default=Z2)
    show_service_details = models.BooleanField(_("show service details"), default=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        self._calculate_total()
        super().save(*args, **kwargs)

    save.alters_data = True

    def _calculate_total(self):
        self.total_excl_tax = (
            self.subtotal - self.discount - getattr(self, "down_payment_total", Z2)
        )
        self.total = self._round_5cents(
            (self.total_excl_tax * (1 + self.tax_rate / 100))
            if self.liable_to_vat
            else self.total_excl_tax
        )

    def _round_5cents(self, value):
        return (value / 5).quantize(Decimal("0.00")) * 5

    @property
    def tax_amount(self):
        return self.total_excl_tax * self.tax_rate / 100 if self.liable_to_vat else Z2

    @property
    def pretty_total_excl(self):
        parts = [gettext("%s excl. tax") % currency(self.total_excl_tax)]
        if self.discount:
            parts.append(" (%s %s)" % (currency(self.discount), gettext("discount")))
        return "".join(parts)



import datetime as dt

from django.db import connections
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.tools.formats import Z1, days, hours


def classify_logging_delay(delay):
    explanation = _("Average logging time is %s after noon.") % hours(delay)

    if delay < 3:
        return _("Promptly"), "success", explanation
    elif delay < 8:
        return _("Same day"), "light", explanation
    elif delay < 30:
        return _("Next day"), "caveat", explanation

    explanation = _("Average logging delay is %s.") % days(delay / 24)
    return _("Late"), "danger", explanation


def mean_logging_delay(date_range):
    with connections["default"].cursor() as cursor:
        cursor.execute(
            """
WITH sq as (
    SELECT
        rendered_by_id AS user_id,
        AVG(created_at - (rendered_on + interval '12 hours')) AS delay
    FROM logbook_loggedhours
    WHERE rendered_on BETWEEN %s AND %s
    GROUP BY user_id
)
SELECT user_id, cast(extract(epoch from delay) as numeric) / 3600
FROM sq
            """,
            date_range,
        )

        return {
            user_id: {"delay": delay, "classification": classify_logging_delay(delay)}
            for user_id, delay in cursor
        }


def logged_hours_stats(date_range):
    with connections["default"].cursor() as cursor:
        cursor.execute(
            """
SELECT rendered_by_id, COUNT(hours), SUM(hours), AVG(hours)
FROM logbook_loggedhours
WHERE rendered_on BETWEEN %s AND %s
GROUP BY rendered_by_id
            """,
            date_range,
        )

        return {
            user_id: {"count": count, "sum": sum, "avg": avg}
            for user_id, count, sum, avg in cursor
        }


def insufficient_breaks(date_range):
    with connections["default"].cursor() as cursor:
        cursor.execute(
            """
with breaks as (
    select
        user_id,
        date_trunc('day', starts_at) as day,
        sum(extract(epoch from (ends_at - starts_at))) as break_seconds
    from logbook_break
    where date_trunc('day', starts_at) between %s and %s
    group by user_id, day
),
logged_hours as (
    select
        rendered_by_id as user_id,
        rendered_on as day,
        sum(hours) as hours
    from logbook_loggedhours
    where rendered_on between %s and %s
    group by user_id, day
),
combined as (
    select
        logged_hours.user_id,
        logged_hours.day,
        hours,
        coalesce(break_seconds, 0) as break_seconds
    from logged_hours
    left join breaks
    on
        breaks.user_id=logged_hours.user_id
        and breaks.day=logged_hours.day
),
insufficient_breaks as (
    select * from combined
    where
        hours >= 9 and break_seconds < 3600
        or hours >= 7 and break_seconds < 1800
        or hours >= 5.5 and break_seconds < 900
)
select
    user_id,
    count(*),
    (
        select count(*)
        from insufficient_breaks
        where insufficient_breaks.user_id=logged_hours.user_id
    )
from logged_hours
group by user_id
            """,
            [*date_range, *date_range],
        )

        return {user_id: {"days": days, "of": of} for user_id, of, days in cursor}


def logbook_stats(date_range):
    mld = mean_logging_delay(date_range)
    lhs = logged_hours_stats(date_range)
    ib = insufficient_breaks(date_range)

    users = [
        {
            "user": user,
            "mean_logging_delay": mld[user.id],
            "logged_hours_stats": lhs[user.id],
            "insufficient_breaks": ib[user.id],
        }
        for user in User.objects.filter(id__in=mld.keys() | lhs.keys())
    ]

    lhs_count = sum((user["logged_hours_stats"]["count"] for user in users), 0)
    lhs_sum = sum((user["logged_hours_stats"]["sum"] for user in users), Z1)

    return {
        "users": users,
        "logged_hours_stats": {
            "count": lhs_count,
            "sum": lhs_sum,
            "avg": lhs_sum / lhs_count if lhs_count else None,
        },
        "insufficient_breaks": {
            "days": sum((user["insufficient_breaks"]["days"] for user in users), 0),
            "of": sum((user["insufficient_breaks"]["of"] for user in users), 0),
        },
    }


def test():  # pragma: no cover
    from pprint import pprint

    today = dt.date.today()
    date_range = [today - dt.timedelta(days=56), today]

    # pprint(mean_logging_delay(date_range))
    # pprint(logged_hours_stats(date_range))
    # pprint(insufficient_breaks(date_range))
    pprint(logbook_stats(date_range))



# Generated by Django 2.2.2 on 2019-06-22 08:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("credit_control", "0005_auto_20190324_1815")]

    operations = [
        migrations.AddField(
            model_name="ledger",
            name="parser",
            field=models.CharField(
                choices=[
                    ("zkb-csv", "ZKB CSV"),
                    ("postfinance-csv", "PostFinance CSV"),
                ],
                default="zkb-csv",
                max_length=20,
                verbose_name="parser",
            ),
            preserve_default=False,
        )
    ]



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext as _, override

from authlib.email import render_to_mail

from workbench.notes.models import Note
from workbench.tools.forms import Form, ModelForm, Textarea


class NoteSearchForm(Form):
    def filter(self, queryset):
        # data = self.cleaned_data
        # queryset = queryset.search(data.get("q"))
        return queryset.select_related("created_by", "content_type")


class NoteForm(ModelForm):
    class Meta:
        model = Note
        fields = ["title", "description"]
        widgets = {"description": Textarea}

    def __init__(self, *args, **kwargs):
        content_object = kwargs.pop("content_object", None)
        super().__init__(*args, **kwargs)

        self.is_new = not self.instance.pk

        if self.is_new:
            self.fields["content_type"] = forms.ModelChoiceField(
                ContentType.objects.all(),
                initial=ContentType.objects.get_for_model(content_object).pk
                if content_object
                else None,
                widget=forms.HiddenInput,
            )
            self.fields["object_id"] = forms.IntegerField(
                initial=content_object.pk if content_object else None,
                widget=forms.HiddenInput,
            )

    def clean(self):
        data = super().clean()
        if self.is_new:
            try:
                data["content_object"] = data["content_type"].get_object_for_this_type(
                    pk=data["object_id"]
                )
            except (KeyError, ObjectDoesNotExist):
                self.add_error(
                    "__all__",
                    _("Unable to determine the object this note should be added to."),
                )
        return data

    def save(self):
        if not self.is_new:
            return super().save()
        instance = super().save(commit=False)
        instance.content_object = self.cleaned_data["content_object"]
        instance.created_by = self.request.user
        instance.save()

        owned_by = getattr(instance.content_object, "owned_by", None)
        if owned_by and owned_by.is_active and owned_by != instance.created_by:
            with override(owned_by.language):
                render_to_mail(
                    "notes/note_notification",
                    {
                        "note": instance,
                        "url": self.request.build_absolute_uri(
                            instance.content_object.get_absolute_url()
                        ),
                    },
                    to=[owned_by.email],
                    reply_to=[owned_by.email, instance.created_by.email],
                ).send(fail_silently=True)

        return instance



import datetime as dt
import io
import os
from decimal import Decimal

from django.conf import settings
from django.test import TestCase

from workbench import factories
from workbench.credit_control.models import CreditEntry
from workbench.credit_control.parsers import (
    parse_postfinance_csv,
    postfinance_preprocess_notice,
    postfinance_reference_number,
)
from workbench.tools.forms import WarningsForm
from workbench.tools.testing import check_code, messages


class CreditEntriesTest(TestCase):
    def test_assignment(self):
        """Batch assignment of credit entries to invoices"""
        for i in range(10):
            invoice = factories.InvoiceFactory.create(
                subtotal=10 + i, liable_to_vat=False
            )

        entry_0 = factories.CreditEntryFactory.create(total=12)
        entry_1 = factories.CreditEntryFactory.create(total=14)
        entry_2 = factories.CreditEntryFactory.create(total=19)

        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/credit-control/assign/")
        # print(response, response.content.decode("utf-8"))

        self.assertContains(response, "widget--radioselect", 3)
        response = self.client.post(
            "/credit-control/assign/",
            {
                "entry_{}_invoice".format(entry_2.pk): invoice.pk,
                "entry_{}_notes".format(entry_1.pk): "Stuff",
            },
        )
        self.assertRedirects(response, "/credit-control/assign/")

        response = self.client.get("/credit-control/assign/")
        self.assertContains(response, "widget--radioselect", 1)

        response = self.client.post(
            "/credit-control/assign/",
            {"entry_{}_notes".format(entry_0.pk): "Stuff"},
            follow=True,
        )
        self.assertRedirects(response, "/credit-control/")
        self.assertEqual(
            messages(response),
            [
                "credit entries have been updated successfully.",
                "All credit entries have already been assigned.",
            ],
        )

    def test_account_statement_upload(self):
        """Uploading account statements with and without duplicates"""
        self.client.force_login(factories.UserFactory.create())
        ledger = factories.LedgerFactory.create()

        def send(data={}):
            with io.open(
                os.path.join(
                    settings.BASE_DIR, "workbench", "test", "account-statement.csv"
                )
            ) as f:
                return self.client.post(
                    "/credit-control/upload/",
                    {"statement": f, "ledger": ledger.pk, **data},
                )

        response = send({"ledger": -1})
        self.assertContains(response, "Select a valid choice.")

        response = send()
        self.assertContains(response, "no-known-payments")

        response = send({WarningsForm.ignore_warnings_id: "no-known-payments"})

        # print(response, response.content.decode("utf-8"))
        self.assertRedirects(response, "/credit-control/")
        self.assertEqual(messages(response), ["Created 2 credit entries."])

        response = send()

        self.assertRedirects(response, "/credit-control/")
        self.assertEqual(messages(response), ["Created 0 credit entries."])

        invoice = factories.InvoiceFactory.create(
            subtotal=Decimal("4000"), _code="00001"
        )
        self.assertAlmostEqual(invoice.total, Decimal("4308.00"))
        response = self.client.get("/credit-control/assign/")
        self.assertContains(response, "<strong><small>00001</small>")

        entry = CreditEntry.objects.get(reference_number="xxxx03130CF54579")
        response = self.client.post(
            entry.urls["update"],
            {
                "ledger": entry.ledger_id,
                "reference_number": entry.reference_number,
                "value_date": entry.value_date,
                "total": entry.total,
                "payment_notice": entry.payment_notice,
                "invoice": invoice.id,
                "notes": "",
            },
        )
        self.assertRedirects(response, entry.urls["detail"])

        invoice.refresh_from_db()
        self.assertEqual(invoice.status, invoice.PAID)
        self.assertEqual(invoice.closed_on, entry.value_date)

    def test_list(self):
        """Filter form validation"""
        self.client.force_login(factories.UserFactory.create())
        ledger = factories.LedgerFactory.create()

        code = check_code(self, "/credit-control/")
        code("")
        code("q=test")
        code("s=pending")
        code("s=processed")
        code("ledger={}".format(ledger.pk))

    def test_create_entry(self):
        """Creating credit entries "by hand" works"""
        self.client.force_login(factories.UserFactory.create())

        response = self.client.post(
            "/credit-control/create/",
            {
                "ledger": factories.LedgerFactory.create().pk,
                "reference_number": "unique",
                "value_date": dt.date.today().isoformat(),
                "total": "20.55",
                "payment_notice": "nothing",
                "notes": "bla",
            },
        )
        self.assertEqual(response.status_code, 302)
        entry = CreditEntry.objects.get()
        self.assertRedirects(response, entry.urls["list"])

    def test_postfinance_utilities(self):
        """Utilities for the PostFinance CSV export"""
        self.assertEqual(
            postfinance_preprocess_notice("bla 2019 -0001-0001 test"),
            "bla 2019-0001-0001 test",
        )
        self.assertEqual(
            postfinance_preprocess_notice("bla 2019-0001-0001 test"),
            "bla 2019-0001-0001 test",
        )
        self.assertEqual(
            postfinance_reference_number(
                "bla bla 190630CH12345678", dt.date(2019, 6, 30)
            ),
            "pf-190630CH12345678",
        )
        self.assertEqual(
            postfinance_reference_number(
                "bla 2019-0001-0001 test", dt.date(2019, 6, 30)
            ),
            "pf-ef8792bffe6303d32130377399828a3f",
        )

    def test_postfinance_parse_csv(self):
        """Generating PostFinance reference numbers when uploading acc. statements"""
        with io.open(
            os.path.join(
                settings.BASE_DIR, "workbench", "test", "postfinance-export.csv"
            ),
            "rb",
        ) as f:
            entries = parse_postfinance_csv(f.read())

        self.assertEqual(len(entries), 3)
        self.assertEqual(entries[0]["total"], Decimal("1193.30"))
        self.assertEqual(entries[0]["reference_number"], "pf-190618CH04D10XYZ")
        self.assertEqual(entries[0]["value_date"], dt.date(2019, 6, 18))

        self.assertEqual(
            entries[1]["reference_number"], "pf-bad7372a51d085b97f7b0e782841490b"
        )
        self.assertIn("2019-0015-0002", entries[1]["payment_notice"])

        self.assertEqual(
            entries[2]["reference_number"], "pf-d0436a6497bf533d2ae49809169f5481"
        )
        self.assertIn("2019-0214-0001", entries[2]["payment_notice"])

    def test_invalid_account_statement(self):
        """Completely invalid account statements do not crash the backend"""
        self.client.force_login(factories.UserFactory.create())
        ledger = factories.LedgerFactory.create()

        with io.open(
            os.path.join(settings.BASE_DIR, "workbench", "test", "stopwatch.png"), "rb"
        ) as f:
            response = self.client.post(
                "/credit-control/upload/", {"statement": f, "ledger": ledger.pk}
            )

        self.assertContains(response, "Error while parsing the statement.")



import datetime as dt

from django import forms
from django.db.models import Q
from django.utils.html import format_html
from django.utils.translation import gettext as _

from workbench.accounts.models import User
from workbench.expenses.models import ExpenseReport
from workbench.logbook.models import LoggedCost
from workbench.tools.formats import currency, local_date_format
from workbench.tools.forms import Form, ModelForm


class ExpenseReportSearchForm(Form):
    s = forms.ChoiceField(
        choices=[
            ("", _("All")),
            (
                _("status"),
                [("in-preparation", _("In preparation")), ("closed", _("Closed"))],
            ),
        ],
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    owned_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["owned_by"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        if data.get("s") == "in-preparation":
            queryset = queryset.filter(closed_on__isnull=True)
        elif data.get("s") == "closed":
            queryset = queryset.filter(closed_on__isnull=False)
        queryset = self.apply_owned_by(queryset)
        return queryset.select_related("owned_by")


class ExpenseReportForm(ModelForm):
    user_fields = default_to_current_user = ["owned_by"]

    class Meta:
        model = ExpenseReport
        fields = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        assert not self.instance.closed_on, "Must be in preparation"
        self.fields["is_closed"] = forms.BooleanField(
            label=_("is closed"),
            help_text=_("Once an expense report is closed it stays that way."),
            required=False,
        )

        if self.instance.pk:
            self.fields["expenses"] = forms.ModelMultipleChoiceField(
                queryset=LoggedCost.objects.expenses(
                    user=self.instance.owned_by
                ).filter(
                    Q(expense_report=self.instance) | Q(expense_report__isnull=True)
                ),
                label=_("expenses"),
                widget=forms.CheckboxSelectMultiple,
            )
            self.fields["expenses"].initial = self.instance.expenses.values_list(
                "id", flat=True
            )
        else:
            self.instance.created_by = self.request.user
            self.instance.owned_by = self.request.user

            expenses = LoggedCost.objects.expenses(user=self.request.user).filter(
                expense_report__isnull=True
            )
            self.fields["expenses"] = forms.ModelMultipleChoiceField(
                queryset=expenses,
                label=_("expenses"),
                widget=forms.CheckboxSelectMultiple,
            )
            self.fields["expenses"].initial = expenses.values_list("id", flat=True)

        self.fields["expenses"].choices = [
            (
                cost.id,
                format_html(
                    "{}<br>{}: {}<br>{}<br>{}{}",
                    local_date_format(cost.rendered_on),
                    cost.service.project,
                    cost.service,
                    cost.description,
                    currency(cost.third_party_costs),
                    " ({} {})".format(cost.expense_currency, cost.expense_cost)
                    if cost.expense_cost
                    else "",
                ),
            )
            for cost in self.fields["expenses"]
            .queryset.select_related("service__project__owned_by")
            .order_by("rendered_on", "pk")
        ]

    def save(self):
        instance = super().save()
        instance.expenses.set(self.cleaned_data["expenses"])
        if self.cleaned_data.get("is_closed"):
            instance.closed_on = dt.date.today()
        instance.save()
        return instance



import datetime as dt
from decimal import Decimal

from django.test import TestCase
from django.urls import reverse
from django.utils import timezone
from django.utils.timezone import localtime
from django.utils.translation import deactivate_all

from time_machine import travel

from workbench import factories
from workbench.accounts.models import User
from workbench.logbook.forms import DetectedTimestampForm
from workbench.timer.models import Timestamp
from workbench.tools.formats import local_date_format


class TimerTest(TestCase):
    def test_timer(self):
        """The timer view does not crash"""
        user = factories.UserFactory.create(is_admin=True)
        self.client.force_login(user)

        response = self.client.get("/timer/")
        self.assertContains(response, "data-current-user")


class TimestampsTest(TestCase):
    def setUp(self):
        deactivate_all()

    @travel("2020-02-20T03:00:00+00:00")
    def test_timestamp(self):
        """Basic smoke test of timestamp creation and deletion"""
        self.client.force_login(factories.UserFactory.create())

        response = self.client.post("/create-timestamp/", {"type": "bla"})
        self.assertEqual(response.status_code, 400)

        response = self.client.post(
            "/create-timestamp/", {"type": "start", "notes": "blub"}
        )
        self.assertEqual(response.status_code, 201)

        t = Timestamp.objects.get()
        self.assertEqual(t.type, "start")
        self.assertEqual(t.notes, "blub")
        self.assertIn("blub", str(t))

        response = self.client.post(reverse("delete_timestamp", args=(t.pk,)))
        self.assertRedirects(response, "/timestamps/")

    def test_timestamp_auth(self):
        """The API supports different authorizations"""
        response = self.client.post("/create-timestamp/", {"type": "bla"})
        self.assertEqual(response.status_code, 400)

        response = self.client.post("/create-timestamp/", {"type": "start"})
        self.assertEqual(response.status_code, 400)

        user = factories.UserFactory.create()

        response = self.client.post(
            "/create-timestamp/", {"type": "start", "user": user.email}
        )
        self.assertEqual(response.status_code, 400)

        response = self.client.post(
            "/create-timestamp/", {"type": "start", "user": user.signed_email}
        )
        self.assertEqual(response.status_code, 201)

        t = Timestamp.objects.get()
        self.assertEqual(t.user, user)

    def test_timestamp_with_time(self):
        """Specifying a time makes the endpoint use it instead of the current time"""
        user = factories.UserFactory.create()

        response = self.client.post(
            "/create-timestamp/",
            {"type": "start", "user": user.signed_email, "time": "09:23"},
        )
        self.assertEqual(response.status_code, 201)

        t = Timestamp.objects.get()
        created_at = timezone.localtime(t.created_at)
        self.assertEqual(created_at.hour, 9)
        self.assertEqual(created_at.minute, 23)

    def test_timestamps_scenario(self):
        """Test a scenario with a break in the middle"""
        today = localtime(timezone.now()).replace(
            hour=9, minute=0, second=0, microsecond=0
        )
        user = factories.UserFactory.create()

        # Insert STOPs at the beginning -- they should be skipped
        ts1 = user.timestamp_set.create(
            type=Timestamp.STOP, created_at=today - dt.timedelta(minutes=60)
        )
        ts2 = user.timestamp_set.create(
            type=Timestamp.STOP, created_at=today - dt.timedelta(minutes=80)
        )

        # t1 =  # unused
        user.timestamp_set.create(
            type=Timestamp.START,
            created_at=today,
            notes="Aaa",
        )
        t2 = user.timestamp_set.create(
            type=Timestamp.STOP,
            created_at=today + dt.timedelta(minutes=40),
            notes="Bbb",
        )
        t3 = user.timestamp_set.create(
            type=Timestamp.STOP, created_at=today + dt.timedelta(minutes=60)
        )
        t4 = user.timestamp_set.create(
            type=Timestamp.STOP, created_at=today + dt.timedelta(minutes=115)
        )
        t5 = user.timestamp_set.create(
            type=Timestamp.STOP, created_at=today + dt.timedelta(minutes=140)
        )
        t6 = user.timestamp_set.create(
            type=Timestamp.STOP, created_at=today + dt.timedelta(minutes=160)
        )

        slices = Timestamp.objects.slices(user)

        partial = [
            (
                slice.elapsed_hours,
                local_date_format(slice.get("starts_at"), fmt="H:i"),
                local_date_format(slice.get("ends_at"), fmt="H:i"),
                slice.get("timestamp_id"),
                slice["description"],
                slice.get("comment"),
            )
            for slice in slices
        ]

        self.assertEqual(
            partial,
            [
                (Decimal(0), "", "07:40", ts2.pk, "", None),
                (Decimal("0.4"), "07:40", "08:00", ts1.pk, "", None),
                (Decimal("1.0"), "08:00", "09:00", None, "", "<detected>"),
                (Decimal("0.7"), "09:00", "09:40", t2.pk, "Aaa; Bbb", None),
                (Decimal("0.4"), "09:40", "10:00", t3.pk, "", None),
                (Decimal("1.0"), "10:00", "10:55", t4.pk, "", None),
                (Decimal("0.5"), "10:55", "11:20", t5.pk, "", None),
                (Decimal("0.4"), "11:20", "11:40", t6.pk, "", None),
            ],
        )

    def test_latest_logbook_entry(self):
        """Timestamps are automatically created for logged hours"""
        now = timezone.now()
        user = factories.UserFactory.create()

        self.assertEqual(Timestamp.objects.slices(user), [])

        user.timestamp_set.create(
            type=Timestamp.START, created_at=now - dt.timedelta(minutes=60)
        )
        l1 = factories.LoggedHoursFactory.create(
            rendered_by=user,
            created_at=now - dt.timedelta(minutes=10),
            description="ABC",
            hours=Decimal("0.2"),  # It was only ten minutes
        )
        user.timestamp_set.create(type=Timestamp.STOP, created_at=now)

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 3)
        # 50 minutes - 2 * 0.1h = 38 minutes ==> 0.7h
        self.assertEqual(slices[0].elapsed_hours, Decimal("0.7"))
        self.assertEqual(slices[1].elapsed_hours, Decimal("0.2"))
        self.assertEqual(slices[2].elapsed_hours, Decimal("0.2"))

        self.assertFalse(slices[0].has_associated_log)
        self.assertTrue(slices[1].has_associated_log)
        self.assertFalse(slices[2].has_associated_log)

        self.assertEqual(l1, slices[1]["description"])

    def test_view(self):
        """The timestamps view does not crash"""
        user = factories.UserFactory.create()
        user.timestamp_set.create(
            type=Timestamp.START, created_at=timezone.now() - dt.timedelta(seconds=9)
        )

        factories.LoggedHoursFactory.create(rendered_by=user)
        factories.BreakFactory.create(user=user)

        self.client.force_login(user)
        response = self.client.get("/timestamps/")
        self.assertContains(response, "timestamps")

    def test_controller(self):
        """The timestamps controller does not crash"""
        response = self.client.get("/timestamps-controller/")
        self.assertContains(response, "Timestamps")

    def test_latest_created_at(self):
        """User.latest_created_at takes logged hours AND timestamps into account"""
        user = factories.UserFactory.create()
        self.assertEqual(user.latest_created_at, None)

        t = user.timestamp_set.create(
            created_at=timezone.now() - dt.timedelta(seconds=99), type=Timestamp.STOP
        )
        user = User.objects.get(id=user.id)
        self.assertEqual(user.latest_created_at, t.created_at)

        h = factories.LoggedHoursFactory.create(rendered_by=user)
        user = User.objects.get(id=user.id)
        self.assertEqual(user.latest_created_at, h.created_at)

        t = user.timestamp_set.create(
            created_at=timezone.now() + dt.timedelta(seconds=99), type=Timestamp.STOP
        )
        user = User.objects.get(id=user.id)
        self.assertEqual(user.latest_created_at, t.created_at)

        h = factories.LoggedHoursFactory.create(
            rendered_by=user, created_at=timezone.now() + dt.timedelta(seconds=300)
        )
        user = User.objects.get(id=user.id)
        self.assertEqual(user.latest_created_at, h.created_at)

        # After assigning the logged_hours relation latest_created_at should
        # be back to Timestamp.created_at
        t.logged_hours = h
        t.save()

        user = User.objects.get(id=user.id)
        self.assertEqual(user.latest_created_at, t.created_at)

    def test_link_timestamps(self):
        """Linking logged hours to timestamps"""
        service = factories.ServiceFactory.create()
        user = service.project.owned_by
        self.client.force_login(user)

        t = user.timestamp_set.create(
            created_at=timezone.now() - dt.timedelta(seconds=99), type=Timestamp.STOP
        )

        slices = Timestamp.objects.slices(user)
        self.assertIn("timestamp={}".format(t.pk), slices[0].hours_create_url)
        self.assertIn("timestamp={}".format(t.pk), slices[0].break_create_url)

        response = self.client.post(
            service.project.urls["createhours"] + "?timestamp={}".format(t.pk),
            {
                "modal-rendered_by": user.pk,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-service": service.id,
                "modal-hours": "0.2",
                "modal-description": "Test",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        t.refresh_from_db()
        self.assertEqual(t.logged_hours.description, "Test")

    def test_list_timestamps(self):
        """The timestamps listing endpoint works"""
        user = factories.UserFactory.create()

        response = self.client.get("/list-timestamps/")
        self.assertEqual(response.status_code, 400)

        response = self.client.get(
            "/list-timestamps/?user={}".format(user.signed_email)
        )
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(
            data, {"hours": "0.0", "user": str(user), "success": True, "timestamps": []}
        )

        user.timestamp_set.create(
            type=Timestamp.START, created_at=timezone.now() - dt.timedelta(seconds=10)
        )
        stop = user.timestamp_set.create(type=Timestamp.STOP)

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 1)
        self.assertEqual(slices[0]["timestamp_id"], stop.id)

        response = self.client.get(
            "/list-timestamps/?user={}".format(user.signed_email)
        )
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(len(data["timestamps"]), 1)
        self.assertEqual(data["timestamps"][0]["elapsed"], "0.1")

    def test_post_split(self):
        """Backwards compatibility: Type "split" still works"""
        self.client.force_login(factories.UserFactory.create())

        response = self.client.post("/create-timestamp/", {"type": "split"})
        self.assertEqual(response.status_code, 201)

        t = Timestamp.objects.get()
        self.assertEqual(t.type, "stop")

    def test_gap_between_hours_and_start(self):
        """A slice is automatically generated if there is a gap between a
        log/stop and a start"""
        hours = factories.LoggedHoursFactory.create(
            created_at=timezone.now() - dt.timedelta(minutes=33),
        )
        user = hours.rendered_by
        user.timestamp_set.create(type=Timestamp.START, notes="Mittag")

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 3)
        self.assertEqual(slices[1]["comment"], "<detected>")

    def test_gap_between_logbook_entries(self):
        """A slice is automatically generated if the gap between subsequent
        logbook entries if there is a gap in-between"""
        user = factories.UserFactory.create()

        factories.LoggedHoursFactory.create(
            rendered_by=user, created_at=timezone.now() - dt.timedelta(seconds=10799)
        )
        factories.LoggedHoursFactory.create(rendered_by=user)

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 3)
        self.assertEqual(slices[1].elapsed_hours, Decimal("2"))

    def test_start_break(self):
        """Start a break, log the break -- further logbook entries should use
        the end of the break as reference time"""
        user = factories.UserFactory.create()

        t = user.timestamp_set.create(
            type=Timestamp.START, created_at=timezone.now() - dt.timedelta(seconds=901)
        )
        t.logged_break = factories.BreakFactory.create(
            user=user, starts_at=t.created_at, ends_at=timezone.now()
        )
        t.save()

        self.assertEqual(len(Timestamp.objects.slices(user)), 1)
        self.assertEqual(user.latest_created_at, t.logged_break.ends_at)

    def test_hours_autocreate_timestamp_from_detected(self):
        """The buttons for <detected> gaps lead to a form which fills the gap
        instead of appending new slices"""
        user = factories.UserFactory.create()

        factories.LoggedHoursFactory.create(
            rendered_by=user, created_at=timezone.now() - dt.timedelta(seconds=10799)
        )
        hours = factories.LoggedHoursFactory.create(rendered_by=user)

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 3)
        self.assertEqual(slices[1].elapsed_hours, Decimal("2"))

        self.assertFalse(slices[1].has_associated_log)
        self.assertIn("detected_ends_at", slices[1].hours_create_url)

        self.client.force_login(user)
        response = self.client.post(
            "{}?{}".format(
                hours.service.project.urls["createhours"],
                slices[1].hours_create_url.split("?")[1],
            ),
            {
                "modal-rendered_by": user.pk,
                "modal-rendered_on": dt.date.today().isoformat(),
                "modal-service": hours.service_id,
                "modal-hours": "2",
                "modal-description": "Test",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 3)  # Still 3

        t = Timestamp.objects.get()
        self.assertEqual(slices[1]["timestamp_id"], t.id)

    def test_break_autocreate_timestamp_from_detected(self):
        """The buttons for <detected> gaps lead to a form which fills the gap
        instead of appending new slices"""
        user = factories.UserFactory.create()

        factories.LoggedHoursFactory.create(
            rendered_by=user, created_at=timezone.now() - dt.timedelta(seconds=10799)
        )
        factories.LoggedHoursFactory.create(rendered_by=user)

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 3)
        self.assertEqual(slices[1].elapsed_hours, Decimal("2"))

        self.assertFalse(slices[1].has_associated_log)
        self.assertIn("detected_ends_at", slices[1].break_create_url)

        self.client.force_login(user)
        response = self.client.post(
            slices[1].break_create_url,
            {
                "modal-user": user.pk,
                "modal-day": dt.date.today().isoformat(),
                "modal-starts_at": localtime(slices[1]["starts_at"]).time().isoformat(),
                "modal-ends_at": localtime(slices[1]["ends_at"]).time().isoformat(),
                "modal-description": "Test",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 3)  # Still 3

        t = Timestamp.objects.get()
        self.assertEqual(slices[1]["timestamp_id"], t.id)

    def test_detected_timestamp_form(self):
        """Test various inputs to the DetectedTimestampForm"""

        self.assertFalse(DetectedTimestampForm({}).is_valid())
        self.assertFalse(DetectedTimestampForm({"detected_ends_at": ""}).is_valid())
        self.assertFalse(
            DetectedTimestampForm({"detected_ends_at": "2010-01-01"}).is_valid()
        )
        self.assertTrue(
            DetectedTimestampForm(
                {"detected_ends_at": "2010-01-01 15:30+02:00"}
            ).is_valid()
        )

    def test_start_with_log(self):
        """Test the behavior of a START timestamp with logged hours"""
        user = factories.UserFactory.create()

        # Create a START 15 minutes ago and a logged hours entry spanning 12 of
        # those 15 minutes
        user.timestamp_set.create(
            type=Timestamp.START,
            created_at=timezone.now() - dt.timedelta(seconds=900),
            logged_hours=factories.LoggedHoursFactory.create(
                hours=Decimal("0.2"),
                created_at=timezone.now() + dt.timedelta(seconds=3600),  # Not now
            ),
        )

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 1)

        # 12 * 60 = 720
        # 900 - 720 ~= 180
        seconds = (timezone.now() - slices[0]["ends_at"]).total_seconds()
        self.assertTrue(175 < seconds < 185)

        # Those three minutes should also be visible here (despite the
        # timestamp created_at value)
        seconds = (timezone.now() - user.latest_created_at).total_seconds()
        self.assertTrue(seconds < 185)

    def test_start_stop_log(self):
        """A start and a stop with a log and still only one slice"""
        user = factories.UserFactory.create()
        t1 = user.timestamp_set.create(
            type=Timestamp.START,
            created_at=timezone.now() - dt.timedelta(seconds=900),
        )
        t2 = user.timestamp_set.create(type=Timestamp.STOP)

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 1)
        self.assertNotEqual(slices[0]["timestamp_id"], t1.pk)
        self.assertEqual(slices[0]["timestamp_id"], t2.pk)
        self.assertEqual(slices[0].elapsed_hours, Decimal("0.3"))

        t2.logged_hours = factories.LoggedHoursFactory.create(
            hours=Decimal("0.3"), rendered_by=user
        )
        t2.save()

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 1)
        self.assertEqual(slices[0]["timestamp_id"], t2.pk)

    def test_elapsed_none_with_only_start(self):
        """A bare START makes one slice with no elapsed time at all"""
        user = factories.UserFactory.create()
        user.timestamp_set.create(
            type=Timestamp.START, created_at=timezone.now() - dt.timedelta(seconds=900)
        )

        slices = Timestamp.objects.slices(user)
        self.assertEqual(len(slices), 1)
        self.assertEqual(slices[0].elapsed_hours, None)



# Generated by Django 2.1.7 on 2019-03-04 21:39

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="last login"
                    ),
                ),
                (
                    "email",
                    models.EmailField(
                        max_length=254, unique=True, verbose_name="email"
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(default=True, verbose_name="is active"),
                ),
                (
                    "is_admin",
                    models.BooleanField(default=False, verbose_name="is admin"),
                ),
                (
                    "_short_name",
                    models.CharField(
                        blank=True, max_length=30, verbose_name="initials"
                    ),
                ),
                (
                    "_full_name",
                    models.CharField(
                        blank=True, max_length=200, verbose_name="full name"
                    ),
                ),
            ],
            options={
                "verbose_name": "user",
                "verbose_name_plural": "users",
                "ordering": ("_full_name",),
            },
        )
    ]



# Generated by Django 2.2.2 on 2019-07-03 06:01

from django.db import migrations, models

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [("contacts", "0006_auto_20190321_1445")]

    operations = [
        migrations.AddField(
            model_name="person",
            name="_fts",
            field=models.TextField(blank=True, editable=False),
        ),
        migrations.RunSQL(
            search.fts(
                "contacts_person",
                ["given_name", "family_name", "address", "notes", "_fts"],
            )
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.core import mail
from django.test import TestCase
from django.utils.translation import deactivate_all

from workbench import factories
from workbench.notes.admin import content_object_url
from workbench.notes.models import Note
from workbench.tools.testing import check_code, messages


class NotesTest(TestCase):
    def setUp(self):
        deactivate_all()

    def test_list(self):
        """Notes listing"""
        deal = factories.DealFactory.create()
        Note.objects.create(content_object=deal, created_by=deal.owned_by, title="Test")
        self.client.force_login(deal.owned_by)

        code = check_code(self, "/notes/")
        code("")

    def test_form(self):
        """Assigning a note to a content object works as expected"""
        deal = factories.DealFactory.create()
        self.client.force_login(deal.owned_by)

        response = self.client.get(deal.urls["detail"])

        content_type = ContentType.objects.get_for_model(deal)

        self.assertContains(response, 'value="{}"'.format(content_type.pk))
        self.assertContains(response, 'value="{}"'.format(deal.pk))

        response = self.client.post(
            "/notes/add-note/",
            {
                "content_type": content_type.pk,
                "object_id": deal.pk,
                "title": "Test title",
                "description": "Test description",
                "next": deal.urls["detail"],
            },
        )
        self.assertRedirects(response, deal.urls["detail"])
        self.assertEqual(messages(response), [])

        response = self.client.post(
            "/notes/add-note/",
            {
                "content_type": content_type.pk,
                "object_id": deal.pk + 1,
                "title": "Test title",
                "description": "Test description",
                "next": deal.urls["detail"],
            },
        )
        self.assertRedirects(response, deal.urls["detail"])
        self.assertEqual(
            messages(response),
            ["Unable to determine the object this note should be added to."],
        )

        # print(response, response.content.decode("utf-8"))

        note = Note.objects.get()
        response = self.client.post(
            note.urls["update"],
            {"title": "Updated", "description": note.description},
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        note.refresh_from_db()
        self.assertEqual(note.title, "Updated")

    def test_note_notification(self):
        """Adding notes to content objects owned by others sends a notification"""
        deal = factories.DealFactory.create()
        self.client.force_login(deal.owned_by)

        content_type = ContentType.objects.get_for_model(deal)

        response = self.client.post(
            "/notes/add-note/",
            {
                "content_type": content_type.pk,
                "object_id": deal.pk,
                "title": "Test title",
                "description": "Test description",
                "next": deal.urls["detail"],
            },
        )
        self.assertRedirects(response, deal.urls["detail"])

        self.assertEqual(len(mail.outbox), 0)

        self.client.force_login(factories.UserFactory.create())
        response = self.client.post(
            "/notes/add-note/",
            {
                "content_type": content_type.pk,
                "object_id": deal.pk,
                "title": "Test title",
                "description": "Test description",
                "next": deal.urls["detail"],
            },
        )
        self.assertRedirects(response, deal.urls["detail"])

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(len(mail.outbox[0].reply_to), 2)

    def test_content_object_url(self):
        """The content_object_url admin helper works"""
        # Has URL
        deal = factories.DealFactory.create()
        content_type = ContentType.objects.get_for_model(deal)

        self.assertEqual(
            content_object_url(Note(content_type=content_type, object_id=42)),
            '<a href="/deals/42/">deal</a>',
        )

        # No URL
        types = factories.service_types()
        content_type = ContentType.objects.get_for_model(types.consulting)

        self.assertEqual(
            content_object_url(Note(content_type=content_type, object_id=42)),
            "service type",
        )



# Generated by Django 2.2.3 on 2019-07-19 07:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("logbook", "0010_auto_20190712_1324"),
        ("logbook", "0010_auto_20190709_2140"),
    ]

    operations = []



from django import forms
from django.conf import settings
from django.contrib import messages
from django.utils.html import format_html
from django.utils.text import capfirst
from django.utils.translation import gettext, gettext_lazy as _, override

from workbench.accounts.features import FEATURES
from workbench.accounts.models import User
from workbench.circles.models import Role
from workbench.contacts.models import Organization, Person
from workbench.projects.models import Campaign, Project, Service
from workbench.services.models import ServiceType
from workbench.tools.forms import Autocomplete, Form, ModelForm, Textarea, add_prefix
from workbench.tools.validation import in_days, is_title_specific


class CampaignSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    s = forms.ChoiceField(
        choices=[
            ("", _("All")),
            (capfirst(_("status")), [("open", _("Open")), ("closed", _("Closed"))]),
        ],
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    org = forms.ModelChoiceField(
        queryset=Organization.objects.all(),
        required=False,
        widget=Autocomplete(model=Organization),
        label="",
    )
    owned_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["owned_by"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        if data.get("s") == "open":
            queryset = queryset.open()
        elif data.get("s") == "closed":
            queryset = queryset.closed()
        queryset = self.apply_renamed(queryset, "org", "customer")
        queryset = self.apply_simple(queryset, "type")
        queryset = self.apply_owned_by(queryset)
        return queryset.select_related("customer", "owned_by")


class CampaignForm(ModelForm):
    user_fields = default_to_current_user = ("owned_by",)

    class Meta:
        model = Campaign
        fields = (
            "customer",
            "title",
            "description",
            "owned_by",
        )
        widgets = {
            "customer": Autocomplete(model=Organization),
            "description": Textarea,
        }

    def __init__(self, *args, **kwargs):
        initial = kwargs.setdefault("initial", {})
        request = kwargs["request"]

        if pk := request.GET.get("copy"):
            try:
                campaign = Campaign.objects.get(pk=pk)
            except (Campaign.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update(
                    {
                        "customer": campaign.customer_id,
                        "title": campaign.title,
                        "description": campaign.description,
                        "owned_by": (
                            campaign.owned_by_id
                            if campaign.owned_by.is_active
                            else request.user.id
                        ),
                    }
                )

        elif pk := request.GET.get("customer"):
            try:
                customer = Organization.objects.get(pk=pk)
            except (Organization.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update({"customer": customer})

        super().__init__(*args, **kwargs)


class CampaignDeleteForm(ModelForm):
    class Meta:
        model = Campaign
        fields = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.projects.exists():
            self.add_warning(
                _(
                    "Projects are linked with this campaign."
                    " They will be released when deleting this campaign."
                ),
                code="release-projects",
            )

    def delete(self):
        self.instance.projects.update(campaign=None)
        self.instance.delete()


class ProjectSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    s = forms.ChoiceField(
        choices=[
            ("all", _("All")),
            (capfirst(_("status")), [("", _("Open")), ("closed", _("Closed"))]),
            (
                _("Defined search"),
                [
                    ("no-invoices", _("No invoices")),
                    ("accepted-offers", _("Accepted offers")),
                    (
                        "accepted-offers-no-invoices",
                        _("Accepted offers but no invoices"),
                    ),
                    ("solely-declined-offers", _("Solely declined offers")),
                    ("old-projects", _("Old projects (60 days inactivity)")),
                    (
                        "invalid-customer-contact-combination",
                        _("Invalid customer/contact combination"),
                    ),
                ],
            ),
        ],
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    org = forms.ModelChoiceField(
        queryset=Organization.objects.all(),
        required=False,
        widget=Autocomplete(model=Organization),
        label="",
    )
    type = forms.ChoiceField(
        choices=[("", _("All types"))] + Project.TYPE_CHOICES,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    owned_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["owned_by"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        if data.get("s") == "":
            queryset = queryset.open()
        elif data.get("s") == "closed":
            queryset = queryset.closed()
        elif data.get("s") == "no-invoices":
            queryset = queryset.orders().without_invoices()
        elif data.get("s") == "accepted-offers":
            queryset = queryset.with_accepted_offers()
        elif data.get("s") == "accepted-offers-no-invoices":
            queryset = queryset.with_accepted_offers().without_invoices()
        elif data.get("s") == "solely-declined-offers":
            queryset = queryset.solely_declined_offers()
        elif data.get("s") == "old-projects":
            queryset = queryset.old_projects()
        elif data.get("s") == "invalid-customer-contact-combination":
            queryset = queryset.invalid_customer_contact_combination()
        queryset = self.apply_renamed(queryset, "org", "customer")
        queryset = self.apply_simple(queryset, "type")
        queryset = self.apply_owned_by(queryset)
        return queryset.select_related("customer", "contact__organization", "owned_by")


class ProjectForm(ModelForm):
    user_fields = default_to_current_user = ("owned_by",)

    class Meta:
        model = Project
        fields = (
            "contact",
            "customer",
            "title",
            "description",
            "campaign",
            "cost_center",
            "owned_by",
            "type",
            "flat_rate",
            "closed_on",
        )
        widgets = {
            "customer": Autocomplete(model=Organization),
            "contact": Autocomplete(model=Person, params={"only_employees": "on"}),
            "campaign": Autocomplete(model=Campaign),
            "description": Textarea,
            "type": forms.RadioSelect,
        }

    def __init__(self, *args, **kwargs):
        initial = kwargs.setdefault("initial", {})
        request = kwargs["request"]

        if pk := request.GET.get("copy"):
            try:
                project = Project.objects.get(pk=pk)
            except (Project.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update(
                    {
                        "customer": project.customer_id,
                        "contact": project.contact_id,
                        "title": project.title,
                        "description": project.description,
                        "type": project.type,
                        "flat_rate": project.flat_rate,
                        "owned_by": (
                            project.owned_by_id
                            if project.owned_by.is_active
                            else request.user.id
                        ),
                    }
                )

        elif pk := request.GET.get("contact"):
            try:
                contact = Person.objects.get(pk=pk)
            except (Person.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update({"customer": contact.organization, "contact": contact})

        elif pk := request.GET.get("customer"):
            try:
                customer = Organization.objects.get(pk=pk)
            except (Organization.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update({"customer": customer})

        for field in ["cost_center", "campaign"]:
            if value := request.GET.get(field):
                initial[field] = value

        super().__init__(*args, **kwargs)
        self.fields["type"].choices = Project.TYPE_CHOICES
        if not self.request.user.features[FEATURES.CONTROLLING]:
            self.fields.pop("flat_rate")
        if not self.request.user.features[FEATURES.CAMPAIGNS]:
            self.fields.pop("campaign")
        if not self.request.user.features[FEATURES.LABOR_COSTS]:
            self.fields.pop("cost_center")
        if self.instance.pk:
            self.fields["closed_on"].help_text = format_html(
                '{} <a href="#" data-field-value="{}">{}</a>',
                _("Set predefined value:"),
                in_days(0).isoformat(),
                _("today"),
            )
        else:
            self.fields.pop("closed_on")

    def clean(self):
        data = super().clean()

        if set(self.changed_data) & {"customer"} and self.instance.invoices.exists():
            self.add_warning(
                _(
                    "This project already has invoices. The invoices'"
                    " customer record will be changed too."
                ),
                code="customer-update-but-already-invoices",
            )

        if (
            set(self.changed_data) & {"flat_rate"}
            and data.get("flat_rate") is not None
            and self.instance.services.exists()
        ):
            self.add_warning(
                _(
                    "You are adding a flat rate to a project which already"
                    " has services. Those services' effort types and rates"
                    " will be overwritten."
                ),
                code="flat-rate-but-already-services",
            )

        return data

    def save(self):
        instance = super().save(commit=False)
        if self.instance.flat_rate is not None:
            with override(settings.WORKBENCH.PDF_LANGUAGE):
                self.instance.services.editable().update(
                    effort_type=gettext("flat rate"),
                    effort_rate=self.instance.flat_rate,
                )
        instance.save()
        if "customer" in self.changed_data:
            instance.invoices.update(customer=instance.customer)
        return instance


class ServiceForm(ModelForm):
    service_type = forms.ModelChoiceField(
        ServiceType.objects.all(),
        label=ServiceType._meta.verbose_name,
        required=False,
        help_text=_("Optional, but useful for quickly filling the fields below."),
    )

    class Meta:
        model = Service
        fields = [
            "title",
            "description",
            "role",
            "allow_logging",
            "offer",
            "is_optional",
            "effort_type",
            "effort_hours",
            "effort_rate",
            "cost",
            "third_party_costs",
        ]
        widgets = {"description": Textarea()}

    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("project", None)
        if not self.project:
            self.project = kwargs["instance"].project
        elif self.project.flat_rate is not None:
            with override(settings.WORKBENCH.PDF_LANGUAGE):
                kwargs["instance"] = Service(
                    effort_type=gettext("flat rate"), effort_rate=self.project.flat_rate
                )

        super().__init__(*args, **kwargs)

        self.fields["role"].choices = Role.objects.choices()
        self.fields["offer"].choices = self.project.offers.in_preparation_choices(
            include=getattr(self.instance, "offer_id", None)
        )
        self.instance.project = self.project

        offer = self.instance.offer
        if offer and offer.status > offer.IN_PREPARATION:
            for field in self.fields:
                if field not in {"role", "allow_logging"}:
                    self.fields[field].disabled = True

            if self.request.method == "GET":
                messages.warning(
                    self.request,
                    _(
                        "Most fields are disabled because service is bound"
                        " to an offer which is not in preparation anymore."
                    ),
                )

        if self.project.flat_rate is not None:
            self.fields["effort_type"].disabled = True
            self.fields["effort_rate"].disabled = True
            self.fields.pop("service_type")

        if not self.request.user.features[FEATURES.GLASSFROG]:
            self.fields.pop("role")

    def clean(self):
        data = super().clean()
        if self.request.user.features[FEATURES.GLASSFROG] and not data.get("role"):
            self.add_warning(_("No role selected."), code="no-role-selected")
        if data.get("title") and not is_title_specific(data["title"]):
            self.add_warning(
                _(
                    "This title seems awfully unspecific."
                    " Please use specific titles for services."
                ),
                code="unspecific-service",
            )
        return data


class ReassignLogbookForm(Form):
    service = forms.ModelChoiceField(
        Service.objects.all(), label=Service._meta.verbose_name
    )

    def __init__(self, *args, **kwargs):
        self.from_service = kwargs.pop("instance")
        super().__init__(*args, **kwargs)
        self.fields[
            "service"
        ].queryset = self.from_service.project.services.logging().exclude(
            pk=self.from_service.pk
        )

        offer = self.from_service.offer
        if not offer or offer.status <= offer.IN_PREPARATION:
            self.fields["try_delete"] = forms.BooleanField(
                label=_("Try deleting the service after reassigning logbook entries"),
                required=False,
            )

    def save(self):
        service = self.cleaned_data["service"]
        self.from_service.loggedhours.update(service=service)
        self.from_service.loggedcosts.update(service=service)

        if self.cleaned_data.get("try_delete") and self.from_service.allow_delete(
            self.from_service, self.request
        ):
            self.from_service.delete()
            messages.success(
                self.request,
                _("%(class)s '%(object)s' has been deleted successfully.")
                % {
                    "class": self.from_service._meta.verbose_name,
                    "object": self.from_service,
                },
            )
        return service


@add_prefix("modal")
class ServiceMoveForm(ModelForm):
    class Meta:
        model = Service
        fields = ["project"]
        widgets = {"project": Autocomplete(model=Project, params={"only_open": "on"})}

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("initial", {}).setdefault("project", "")
        super().__init__(*args, **kwargs)

    def clean(self):
        data = super().clean()
        if data.get("project") and data.get("project").closed_on:
            self.add_error("project", _("This project is already closed."))

        if (
            data.get("project")
            and self.instance.project.flat_rate is None
            and data["project"].flat_rate is not None
        ):
            self.add_warning(
                _(
                    "The project %(project)s has a flat rate which will be applied"
                    " to this service too."
                )
                % {"project": data["project"]},
                code="new-project-has-flat-rate",
            )
        return data

    def save(self):
        instance = super().save(commit=False)
        if instance.project.flat_rate is not None:
            instance.effort_type = gettext("flat rate")
            instance.effort_rate = instance.project.flat_rate
        instance.save()
        return instance


@add_prefix("modal")
class ProjectAutocompleteForm(forms.Form):
    project = forms.ModelChoiceField(
        queryset=Project.objects.all(),
        widget=Autocomplete(model=Project, params={"only_open": "on"}),
        label="",
        required=False,
    )
    service = forms.ModelChoiceField(
        queryset=Service.objects.all(),
        widget=Autocomplete(model=Service),
        label="",
        required=False,
    )

    def clean(self):
        data = super().clean()
        if data["service"]:
            data["project"] = data["service"].project
            return data
        elif data["project"]:
            return data
        self.add_error("project", _("This field is required."))
        return data


class OffersRenumberForm(Form):
    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("instance")

        super().__init__(*args, **kwargs)

        self.offers = self.project.offers.order_by("_code")
        for offer in self.offers:
            self.fields["offer_{}_code".format(offer.pk)] = forms.IntegerField(
                label=str(offer),
                initial=offer._code,
                min_value=1,
            )

    def clean(self):
        data = super().clean()
        self.codes = {
            offer: data.get("offer_{}_code".format(offer.pk)) for offer in self.offers
        }
        if len(self.codes) != len(set(self.codes.values())):
            raise forms.ValidationError(_("Codes must be unique."))
        return data

    def save(self):
        for offer, code in self.codes.items():
            offer._code = code
            offer.save()
        return self.project



from workbench.tools import admin

from . import models


@admin.register(models.ExpenseReport)
class ExpenseReportAdmin(admin.ModelAdmin):
    list_display = ("created_at", "created_by", "owned_by", "total")
    raw_id_fields = ("created_by", "owned_by")


@admin.register(models.ExchangeRates)
class ExchangeRatesAdmin(admin.ModelAdmin):
    pass



from django.conf import settings
from django.contrib import auth, messages
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.translation import gettext as _, gettext_lazy
from django.views.decorators.cache import never_cache

from authlib.google import GoogleOAuth2Client

from workbench.accounts.forms import UserForm
from workbench.accounts.models import User
from workbench.accounts.reporting import logged_hours
from workbench.generic import DetailView, UpdateView


def accounts(request):
    return redirect("accounts_update" if request.user.is_authenticated else "login")


class UserUpdateView(UpdateView):
    model = User
    form_class = UserForm
    success_url = "/"
    title = gettext_lazy("Settings")

    def get_object(self):
        if self.request.user.is_authenticated:
            return self.request.user
        elif "user_email" in self.request.session:
            email = self.request.session["user_email"]
            return User(email=email, _short_name=email.split("@")[0])
        else:
            raise Http404

    def form_valid(self, form):
        response = HttpResponseRedirect(self.get_success_url())
        self.object = form.save(commit=False)
        if self.object.pk:  # Existed already
            self.object.save()
            return response

        self.object.email = self.request.session.pop("user_email")
        self.object.save()
        auth.login(self.request, auth.authenticate(email=self.object.email))
        response.set_cookie("login_hint", self.object.email, expires=180 * 86400)
        return response


@never_cache
def login(request):
    if request.user.is_authenticated:
        return redirect("/")
    return render(
        request,
        "accounts/login.html",
        {
            "auth_params": "?login_hint=&prompt=consent+select_account",
            "auth_button": _("Select Google account"),
        }
        if request.GET.get("error")
        else {},
    )


@never_cache
def oauth2(request):
    auth_params = request.GET.dict()
    auth_params.setdefault("login_hint", request.COOKIES.get("login_hint", ""))
    client = GoogleOAuth2Client(request, authorization_params=auth_params)

    if not request.GET.get("code"):
        return redirect(client.get_authentication_url())

    try:
        user_data = client.get_user_data()
    except Exception:
        messages.error(request, _("Error while fetching user data. Please try again."))
        return redirect("login")

    email = user_data.get("email")
    user = auth.authenticate(request, email=email)
    if user and user.is_active:
        auth.login(request, user)
        next = request.get_signed_cookie("next", default=None, salt="next")
        response = redirect(next if next else "/")
        response.delete_cookie("next")
        response.set_cookie("login_hint", user.email, expires=180 * 86400)
        return response

    elif User.objects.filter(email=email).exists():
        messages.error(
            request, _("The user with email address %s is inactive.") % email
        )
        response = HttpResponseRedirect("{}?error=1".format(reverse("login")))
        response.delete_cookie("login_hint")
        return response

    elif email.endswith("@%s" % settings.WORKBENCH.SSO_DOMAIN):
        messages.info(request, _("Welcome! Please fill in your details."))
        request.session["user_email"] = email
        return redirect("accounts_update")

    else:
        messages.error(request, _("No user with email address %s found.") % email)
        response = HttpResponseRedirect("{}?error=1".format(reverse("login")))
        response.delete_cookie("login_hint")
        return response


def logout(request):
    auth.logout(request)
    messages.success(request, _("You have been signed out."))
    response = redirect("login")
    response.delete_cookie("login_hint")
    return response


class ProfileView(DetailView):
    model = User
    queryset = User.objects.active()
    template_name_suffix = "_statistics"

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            logged_hours=logged_hours(self.object), **kwargs
        )



# Generated by Django 2.1.7 on 2019-03-12 13:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("accounts", "0002_auto_20190304_2247")]

    operations = [
        migrations.AddField(
            model_name="user",
            name="enforce_same_week_logging",
            field=models.BooleanField(
                default=True, verbose_name="enforce same week logging"
            ),
        )
    ]



from django.contrib.messages import get_messages


def messages(response):
    return [m.message for m in get_messages(response.wsgi_request)]


def check_code(test, base_url):
    def code(param, status_code=200):
        response = test.client.get("{}?{}".format(base_url, param))
        test.assertEqual(response.status_code, status_code)

    return code



from django.utils.translation import gettext_lazy as _

from admin_ordering.admin import OrderableAdmin

from workbench.tools import admin

from . import models


@admin.register(models.WorkingTimeModel)
class WorkingTimeModelAdmin(OrderableAdmin, admin.ReadWriteModelAdmin):
    list_display = ["name", "position"]
    list_editable = ["position"]
    ordering_field = "position"


@admin.register(models.Year)
class YearAdmin(admin.ReadWriteModelAdmin):
    list_display = [
        "year",
        "working_time_model",
        "working_time_per_day",
        *models.Year.MONTHS,
        "days",
    ]

    def days(self, instance):
        return sum(instance.months)

    days.short_description = _("days")


@admin.register(models.Employment)
class EmploymentAdmin(admin.ReadWriteModelAdmin):
    list_display = [
        "user",
        "date_from",
        "date_until",
        "percentage",
        "vacation_weeks",
        "notes",
    ]
    list_filter = ["user", "date_from"]


@admin.register(models.Absence)
class AbsenceAdmin(admin.ModelAdmin):
    list_display = ["user", "starts_on", "days", "description", "is_vacation"]
    radio_fields = {"reason": admin.HORIZONTAL}
    readonly_fields = ["is_vacation"]



import datetime as dt
from collections import defaultdict

from django.db.models import Prefetch
from django.utils import timezone

from workbench.audit.models import LoggedAction
from workbench.deals.models import Deal, Value, ValueType
from workbench.tools.formats import Z2
from workbench.tools.history import EVERYTHING, changes


def accepted_deals(date_range, *, users=None):
    queryset = (
        Deal.objects.filter(status=Deal.ACCEPTED, closed_on__range=date_range)
        .select_related("customer", "contact__organization", "owned_by")
        .prefetch_related(
            Prefetch("values", queryset=Value.objects.select_related("type"))
        )
        .order_by("closed_on")
    )

    if users is not None:
        queryset = queryset.filter(owned_by__in=users)

    by_user = defaultdict(list)
    by_month_and_valuetype = defaultdict(lambda: defaultdict(lambda: Z2))

    valuetypes = set()

    for deal in queryset:
        by_user[deal.owned_by].append(deal)

        month = deal.closed_on.replace(day=1)
        for value in deal.values.all():
            by_month_and_valuetype[month][value.type] += value.value
            valuetypes.add(value.type)

    deals = [
        {
            "user": user,
            "user_id": str(user.id),
            "deals": deals,
            "count": len(deals),
            "sum": sum(deal.value for deal in deals),
        }
        for user, deals in by_user.items()
    ]

    valuetypes |= set(ValueType.objects.filter(weekly_target__isnull=False))
    valuetypes = sorted(valuetypes)
    months = [
        {
            "month": month,
            "sum": sum(by_valuetype.values()),
            "values": [
                {"type": type, "value": by_valuetype.get(type, Z2)}
                for type in valuetypes
            ],
        }
        for month, by_valuetype in by_month_and_valuetype.items()
    ]
    date_range_length = (date_range[1] - date_range[0]).days + 1

    return {
        "by_user": sorted(deals, key=lambda row: row["sum"], reverse=True),
        "by_month_and_valuetype": sorted(months, key=lambda row: row["month"]),
        "by_valuetype": [
            {
                "type": type,
                "sum": sum(
                    by_valuetype[type]
                    for by_valuetype in by_month_and_valuetype.values()
                ),
                "target": type.weekly_target * date_range_length / 7
                if type.weekly_target is not None
                else None,
            }
            for type in valuetypes
        ],
        "valuetypes": valuetypes,
        "sum": sum(row["sum"] for row in deals),
        "count": sum(row["count"] for row in deals),
        "target": sum(type.weekly_target or 0 for type in valuetypes)
        * date_range_length
        / 7,
        "weeks": date_range_length / 7,
    }


def declined_deals(date_range, *, users=None):
    queryset = (
        Deal.objects.filter(status=Deal.DECLINED, closed_on__range=date_range)
        .select_related("customer", "contact__organization", "owned_by")
        .prefetch_related(
            Prefetch("values", queryset=Value.objects.select_related("type"))
        )
        .order_by("-closed_on")
    )
    return queryset


def deal_history(date_range, *, users=None):
    actions = LoggedAction.objects.for_model(Deal).filter(
        created_at__range=[
            timezone.make_aware(dt.datetime.combine(day, time))
            for day, time in zip(date_range, [dt.time.min, dt.time.max])
        ]
    )
    fields = {
        "id",
        "title",
        "value",
        "probability",
        "decision_expected_on",
        "status",
        "closing_type",
        "closing_notice",
    }
    user_ids = EVERYTHING if users is None else {user.id for user in users} | {None}
    return [
        change
        for change in changes(Deal, fields, actions)
        if change.version.user_id in user_ids
    ]


def test():  # pragma: no cover
    from pprint import pprint

    pprint(accepted_deals([dt.date(2020, 1, 1), dt.date(2020, 3, 31)]))



# Generated by Django 3.1a1 on 2020-06-21 18:50

from decimal import Decimal

import django.contrib.postgres.fields
import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("offers", "0009_offer_valid_until"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("projects", "0019_auto_20200523_0929"),
    ]

    operations = [
        migrations.CreateModel(
            name="PlanningTeamMembership",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="planning_team_memberships",
                        to="projects.project",
                        verbose_name="project",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="planning_team_memberships",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "planning team membership",
                "verbose_name_plural": "planning team memberships",
            },
        ),
        migrations.CreateModel(
            name="PlanningRequest",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "requested_hours",
                    models.DecimalField(
                        decimal_places=1,
                        max_digits=10,
                        validators=[
                            django.core.validators.MinValueValidator(Decimal("0.1"))
                        ],
                        verbose_name="requested hours",
                    ),
                ),
                (
                    "planned_hours",
                    models.DecimalField(
                        decimal_places=1,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="planned hours",
                    ),
                ),
                (
                    "earliest_start_on",
                    models.DateField(verbose_name="earliest start on"),
                ),
                (
                    "completion_requested_on",
                    models.DateField(verbose_name="completion requested on"),
                ),
                ("title", models.CharField(max_length=100, verbose_name="title")),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="description"),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "closed_at",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="closed at"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="sent_planning_requests",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="created by",
                    ),
                ),
                (
                    "offer",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="planning_requests",
                        to="offers.offer",
                        verbose_name="offer",
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="planning_requests",
                        to="projects.project",
                        verbose_name="project",
                    ),
                ),
                (
                    "receivers",
                    models.ManyToManyField(
                        related_name="received_planning_requests",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="receivers",
                    ),
                ),
            ],
            options={
                "verbose_name": "planning request",
                "verbose_name_plural": "planning requests",
                "ordering": ["-pk"],
            },
        ),
        migrations.CreateModel(
            name="PlannedWork",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "planned_hours",
                    models.DecimalField(
                        decimal_places=1,
                        max_digits=10,
                        validators=[
                            django.core.validators.MinValueValidator(Decimal("0.1"))
                        ],
                        verbose_name="planned hours",
                    ),
                ),
                ("notes", models.TextField(blank=True, verbose_name="notes")),
                (
                    "weeks",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.DateField(), size=None, verbose_name="weeks"
                    ),
                ),
                (
                    "offer",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="planned_work",
                        to="offers.offer",
                        verbose_name="offer",
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="planned_work",
                        to="projects.project",
                        verbose_name="project",
                    ),
                ),
                (
                    "request",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="planned_work",
                        to="planning.planningrequest",
                        verbose_name="planning request",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="planned_work",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "planned work",
                "verbose_name_plural": "planned work",
                "ordering": ["-pk"],
            },
        ),
        migrations.RunSQL(
            "SELECT audit_audit_table('planning_planningrequest');"
            "SELECT audit_audit_table('planning_plannedwork');"
        ),
    ]



# Generated by Django 3.0.4 on 2020-03-10 18:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("reporting", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="CostCenter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "position",
                    models.PositiveIntegerField(default=0, verbose_name="position"),
                ),
            ],
            options={
                "verbose_name": "cost center",
                "verbose_name_plural": "cost centers",
                "ordering": ("position", "id"),
            },
        ),
        migrations.RunSQL(
            "SELECT audit_audit_table('reporting_costcenter', ARRAY['position']);"
        ),
    ]



# Generated by Django 2.2.1 on 2019-05-18 12:43

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("logbook", "0007_auto_20190518_1425")]

    operations = [
        migrations.RunSQL("UPDATE logbook_loggedcost SET rendered_by_id=created_by_id")
    ]



from django.urls import re_path

from workbench import generic
from workbench.notes import views
from workbench.notes.forms import NoteForm, NoteSearchForm
from workbench.notes.models import Note


urlpatterns = [
    re_path(r"^add-note/$", views.add_note, name="notes_note_add"),
    re_path(
        r"^$",
        generic.ListView.as_view(
            model=Note, search_form_class=NoteSearchForm, show_create_button=False
        ),
        name="notes_note_list",
    ),
    re_path(
        r"(?P<pk>[0-9]+)/update/$",
        generic.UpdateView.as_view(model=Note, form_class=NoteForm),
        name="notes_note_update",
    ),
    re_path(
        r"^(?P<pk>[0-9]+)/delete/$",
        generic.DeleteView.as_view(
            model=Note, template_name="modal_confirm_delete.html"
        ),
        name="notes_note_delete",
    ),
]



# Generated by Django 2.1.7 on 2019-03-12 11:54

from decimal import Decimal

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("logbook", "0004_auto_20190310_1134")]

    operations = [
        migrations.AlterField(
            model_name="loggedhours",
            name="hours",
            field=models.DecimalField(
                decimal_places=1,
                max_digits=10,
                validators=[django.core.validators.MinValueValidator(Decimal("0.1"))],
                verbose_name="hours",
            ),
        )
    ]



# Generated by Django 3.0 on 2019-12-07 12:04

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("awt", "0005_auto_20191207_1304"),
        ("accounts", "0004_user_language"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="working_time_model",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="awt.WorkingTimeModel",
                verbose_name="working time model",
            ),
        ),
    ]



from django.db import models
from django.db.models import Max
from django.utils import timezone
from django.utils.text import Truncator
from django.utils.translation import gettext_lazy as _

from workbench.tools.formats import Z1, Z2
from workbench.tools.models import HoursFieldAllowNegatives, Model, MoneyField
from workbench.tools.validation import raise_if_errors


class ServiceType(Model):
    title = models.CharField(_("title"), max_length=40)
    hourly_rate = MoneyField(_("hourly rate"))
    position = models.IntegerField(_("position"), default=0)

    class Meta:
        ordering = ("position", "id")
        verbose_name = _("service type")
        verbose_name_plural = _("service types")

    def __str__(self):
        return self.title


class ServiceBase(Model):
    created_at = models.DateTimeField(_("created at"), default=timezone.now)

    title = models.CharField(_("title"), max_length=200)
    description = models.TextField(_("description"), blank=True)
    position = models.IntegerField(_("position"), default=0)

    service_hours = HoursFieldAllowNegatives(_("service hours"), default=0)
    service_cost = MoneyField(_("service cost"), default=0)

    effort_type = models.CharField(_("effort type"), max_length=50, blank=True)
    effort_hours = HoursFieldAllowNegatives(_("hours"), blank=True, null=True)
    effort_rate = MoneyField(_("hourly rate"), blank=True, null=True)

    cost = MoneyField(_("cost"), blank=True, null=True)
    third_party_costs = MoneyField(
        _("third party costs"),
        default=None,
        blank=True,
        null=True,
        help_text=_("Total incl. tax for third-party services."),
    )

    class Meta:
        abstract = True
        ordering = ["position", "created_at"]
        verbose_name = _("service")
        verbose_name_plural = _("services")

    def __str__(self):
        return Truncator(": ".join(filter(None, (self.title, self.description)))).chars(
            100
        )

    def project_service_title(self):
        title = "{}: {}{}{}".format(
            self.project,
            self.title,
            ": " if self.description else "",
            self.description,
        )
        return Truncator(title).chars(100)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._related_model = self._meta.get_field(self.RELATED_MODEL_FIELD)
        self._orig_related_id = getattr(self, self._related_model.attname)

    def save(self, *args, **kwargs):
        skip_related_model = kwargs.pop("skip_related_model", False)

        if not self.position:
            max_pos = self.__class__._default_manager.aggregate(m=Max("position"))["m"]
            self.position = 10 + (max_pos or 0)
        self.service_hours = self.effort_hours or Z1
        self.service_cost = self.cost or Z2
        if all((self.effort_hours, self.effort_rate)):
            self.service_cost += self.effort_hours * self.effort_rate

        super().save(*args, **kwargs)

        if not skip_related_model:
            ids = filter(
                None,
                [self._orig_related_id, getattr(self, self._related_model.attname)],
            )
            for (
                instance
            ) in self._related_model.remote_field.model._default_manager.filter(
                id__in=ids
            ):
                instance.save()

    save.alters_data = True

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        if self._orig_related_id:
            ids = filter(
                None,
                [self._orig_related_id, getattr(self, self._related_model.attname)],
            )
            for (
                instance
            ) in self._related_model.remote_field.model._default_manager.filter(
                id__in=ids
            ):
                instance.save()

    delete.alters_data = True

    def clean_fields(self, exclude):
        super().clean_fields(exclude)
        errors = {}
        effort = (self.effort_type != "", self.effort_rate is not None)
        if any(effort) and not all(effort):
            if self.effort_type == "":
                errors["effort_type"] = _("Either fill in all fields or none.")
            if self.effort_rate is None:
                errors["effort_rate"] = _("Either fill in all fields or none.")
        if self.third_party_costs is not None and self.cost is None:
            errors["cost"] = _("Cannot be empty if third party costs is set.")
        raise_if_errors(errors, exclude)



from workbench.tools import admin

from . import models


@admin.register(models.Service)
class ServiceAdmin(admin.ModelAdmin):
    list_display = (
        "title",
        "project",
        "offer",
        "service_hours",
        "service_cost",
    )
    list_select_related = ["project__owned_by", "offer__project", "offer__owned_by"]
    ordering = ["-pk"]
    raw_id_fields = ["project", "offer", "role"]
    search_fields = ["project__title", "title", "description"]


@admin.register(models.Project)
class ProjectAdmin(admin.ModelAdmin):
    list_display = ("title", "customer", "owned_by", "type", "closed_on")
    list_filter = ["type", "closed_on"]
    raw_id_fields = ["customer", "contact", "owned_by", "campaign"]


@admin.register(models.Campaign)
class CampaignAdmin(admin.ModelAdmin):
    list_display = ["title", "customer", "owned_by"]
    raw_id_fields = ["customer", "owned_by"]



# Generated by Django 2.1.7 on 2019-03-21 15:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("audit", "0001_initial"),
        ("credit_control", "0003_auto_20190321_1447"),
        ("logbook", "0001_initial"),
    ]

    operations = [
        migrations.RunSQL(
            "UPDATE audit_logged_actions SET row_data=delete(row_data, 'fts_document');"
            "UPDATE audit_logged_actions SET changed_fields=delete(changed_fields, 'fts_document');"
        ),
        migrations.RunSQL(
            "SELECT audit_audit_table('contacts_organization', ARRAY['fts_document']);"
            "SELECT audit_audit_table('contacts_person', ARRAY['fts_document']);"
            "SELECT audit_audit_table('credit_control_creditentry', ARRAY['fts_document']);"
            "SELECT audit_audit_table('invoices_invoice', ARRAY['fts_document']);"
            "SELECT audit_audit_table('invoices_recurringinvoice', ARRAY['fts_document']);"
            "SELECT audit_audit_table('logbook_loggedhours', ARRAY['fts_document']);"
            "SELECT audit_audit_table('logbook_loggedcost', ARRAY['fts_document']);"
            "SELECT audit_audit_table('offers_offer', ARRAY['fts_document']);"
            "SELECT audit_audit_table('projects_project', ARRAY['fts_document']);"
            "UPDATE contacts_organization SET id=id;"
            "UPDATE contacts_person SET id=id;"
            "UPDATE credit_control_creditentry SET id=id;"
            "UPDATE invoices_invoice SET id=id;"
            "UPDATE invoices_recurringinvoice SET id=id;"
            "UPDATE logbook_loggedhours SET id=id;"
            "UPDATE logbook_loggedcost SET id=id;"
            "UPDATE offers_offer SET id=id;"
            "UPDATE projects_project SET id=id;"
        ),
    ]



import datetime as dt

from django.apps import apps
from django.contrib import messages
from django.db import connections
from django.http import Http404
from django.shortcuts import render
from django.urls import reverse
from django.utils.text import capfirst
from django.utils.translation import gettext as _

from workbench.accounts.features import FEATURES
from workbench.audit.models import LoggedAction
from workbench.contacts.models import Organization, Person
from workbench.deals.models import Deal
from workbench.invoices.models import Invoice, RecurringInvoice
from workbench.logbook.models import LoggedHours
from workbench.offers.models import Offer
from workbench.planning.models import PlanningRequest
from workbench.projects.models import Campaign, Project
from workbench.tools.history import HISTORY, changes
from workbench.tools.validation import in_days


def _needs_action(user):
    rows = []
    if user.features[FEATURES.DEALS]:
        rows.append(
            {
                "type": "deals",
                "verbose_name_plural": Deal._meta.verbose_name_plural,
                "url": Deal.urls["list"],
                "objects": Deal.objects.maybe_actionable(user=user),
            }
        )
    if user.features[FEATURES.CONTROLLING]:
        rows.append(
            {
                "type": "offers",
                "verbose_name_plural": Offer._meta.verbose_name_plural,
                "url": Offer.urls["list"],
                "objects": Offer.objects.maybe_actionable(user=user),
            }
        )
    if user.features[FEATURES.PLANNING]:
        rows.append(
            {
                "type": "planninrequests",
                "verbose_name_plural": PlanningRequest._meta.verbose_name_plural,
                "url": user.urls["planning"],
                "objects": PlanningRequest.objects.maybe_actionable(user=user),
            }
        )

    return [row for row in rows if row["objects"]]


def _todays_hours(user):
    return (
        LoggedHours.objects.filter(rendered_by=user, rendered_on=dt.date.today())
        .select_related("service__project__owned_by", "rendered_by", "timestamp")
        .order_by("-created_at")[:15]
    )


def _all_users_hours():
    return (
        LoggedHours.objects.filter(rendered_on__gte=in_days(-7))
        .select_related("service__project__owned_by", "rendered_by", "timestamp")
        .order_by("-created_at")[:20]
    )


def _birthdays():
    with connections["default"].cursor() as cursor:
        cursor.execute(
            """
SELECT id, given_name, family_name, date_of_birth, is_active_user FROM (
    SELECT
        p.id,
        given_name,
        family_name,
        date_of_birth,
        u.is_active AS is_active_user,
        (current_date - date_of_birth) % 365.24 AS diff
    FROM contacts_person p
    LEFT OUTER JOIN accounts_user u ON p.id=u.person_id
    WHERE date_of_birth is not null AND is_archived=FALSE
) AS subquery
WHERE diff < 7 or diff > 350
ORDER BY (diff + 180) % 365 DESC
            """
        )
        return [
            {
                "id": row[0],
                "given_name": row[1],
                "family_name": row[2],
                "date_of_birth": row[3],
                "is_active_user": row[4],
            }
            for row in cursor
        ]


def _in_preparation(user):
    invoices = Invoice.objects.filter(
        owned_by=user, status=Invoice.IN_PREPARATION
    ).select_related("project", "owned_by")
    offers = Offer.objects.filter(
        owned_by=user, status=Offer.IN_PREPARATION
    ).select_related("project", "owned_by")

    return {
        key: value
        for key, value in [("invoices", invoices), ("offers", offers)]
        if value
    }


def start(request):
    request.user.take_a_break_warning(request=request)

    return render(
        request,
        "start.html",
        {
            "needs_action": _needs_action(request.user),
            "todays_hours": _todays_hours(request.user),
            "all_users_hours": _all_users_hours(),
            "birthdays": _birthdays(),
            "in_preparation": _in_preparation(request.user),
        },
    )


def search(request):
    results = []
    if q := request.GET.get("q", ""):
        sources = [
            Project.objects.select_related("owned_by"),
        ]
        if request.user.features[FEATURES.CAMPAIGNS]:
            sources.append(Campaign.objects.select_related("owned_by"))
        sources.extend(
            [
                Organization.objects.active(),
                Person.objects.active().select_related("organization"),
            ]
        )
        if request.user.features[FEATURES.CONTROLLING]:
            sources.extend(
                [
                    Invoice.objects.select_related("project", "owned_by"),
                    RecurringInvoice.objects.all(),
                    Offer.objects.select_related("project", "owned_by").order_by("-pk"),
                    Deal.objects.order_by("-pk"),
                ]
            )
        results = [
            {
                "verbose_name_plural": queryset.model._meta.verbose_name_plural,
                "url": reverse(
                    "%s_%s_list"
                    % (queryset.model._meta.app_label, queryset.model._meta.model_name)
                ),
                "results": queryset.search(q)[:101],
            }
            for queryset in sources
        ]
    else:
        messages.error(request, _("Search query missing."))

    return render(request, "search.html", {"query": q, "results": results})


DB_TABLE_TO_MODEL = {model._meta.db_table: model for model in apps.get_models()}


def history(request, db_table, attribute, id):
    try:
        model = DB_TABLE_TO_MODEL[db_table]
        cfg = HISTORY[model]
    except KeyError:
        raise Http404

    if callable(cfg):
        cfg = cfg(request.user)
    fields = cfg.get("fields", set())

    instance = None
    title = None
    related = []

    if attribute == "id":
        try:
            instance = model._base_manager.get(**{attribute: id})
        except Exception:
            instance = None
            title = model._meta.verbose_name

        related = [
            (
                capfirst(model._meta.verbose_name_plural),
                reverse("history", args=(model._meta.db_table, attribute, id)),
            )
            for model, attribute in cfg.get("related", [])
        ]
    else:
        title = _("%(model)s with %(attribute)s=%(id)s") % {
            "model": capfirst(model._meta.verbose_name_plural),
            "attribute": attribute,
            "id": id,
        }

    actions = LoggedAction.objects.for_model(model).with_data(**{attribute: id})

    return render(
        request,
        "history_modal.html",
        {
            "instance": instance,
            "title": title,
            "changes": changes(model, fields, actions),
            "related": related,
        },
    )
